import logging

import pandas as pd
from sam.logging_functions import log_dataframe_characteristics

logger = logging.getLogger(__name__)


def top_n_correlations(
    df: pd.DataFrame,
    goal_feature: str,
    n: int = 5,
    grouped: bool = True,
    sep: str = "#",
):
    """
    Given a dataset, retrieve the top n absolute correlating features per group or in general

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe containing the features that have to be correlated.
    goal_feature: str
        Feature that is used to compare the correlation with other features
    n: int (default: 5)
        Number of correlating features that are returned
    grouped: bool (default: True)
        Whether to group the features and take the top n of a group, or just the top n
        correlating features in general. Groups are created based on column name, and
        are all characters before the first occurence of the sep. s
        For example, if the sep is '#', then DEBIET_TOTAAL#lag_0 is in group DEBIET_TOTAAL
    sep: str (default: '#')
        The seperator character. The group of a column is defined as everything before the first
        occurence of this character. Only relevant if grouped is True

    Returns
    -------
    df: pd.DataFrame
        If grouped is true, a dataframe containing 3 columns (GROUP, index, goal_variable) is
        returned, else a dataframe containing 2 columns (index, goal_variable) is returned.
        index contains the correlating features and goal_variable the correlation value.
        GROUP contains the group.


    Examples
    --------
    >>> import pandas as pd
    >>> from sam.feature_engineering import BuildRollingFeatures
    >>> from sam.exploration import top_n_correlations
    >>> import numpy as np
    >>> goal_feature = 'DEBIET_TOTAAL#lag_0'
    >>> df = pd.DataFrame({
    ...                'RAIN': [0.1, 0.2, 0.0, 0.6, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ...                'DEBIET_A': [1, 2, 3, 4, 5, 5, 4, 3, 2, 4, 2, 3],
    ...                'DEBIET_B': [3, 1, 2, 3, 3, 6, 4, 1, 3, 3, 1, 5]})
    >>> df['DEBIET_TOTAAL'] = df['DEBIET_A'] + df['DEBIET_B']
    >>> RollingFeatures = BuildRollingFeatures(rolling_type='lag',
    ...     window_size = np.arange(12), lookback=0, keep_original=False)
    >>> res = RollingFeatures.fit_transform(df)
    >>> top_n_correlations(res, goal_feature, n=2, grouped=True, sep='#')
               GROUP                 index  DEBIET_TOTAAL#lag_0
    0       DEBIET_A       DEBIET_A#lag_10             1.000000
    1       DEBIET_A        DEBIET_A#lag_0             0.838591
    2       DEBIET_B       DEBIET_B#lag_10            -1.000000
    3       DEBIET_B        DEBIET_B#lag_0             0.897340
    4  DEBIET_TOTAAL  DEBIET_TOTAAL#lag_10            -1.000000
    5  DEBIET_TOTAAL   DEBIET_TOTAAL#lag_9             0.944911
    6           RAIN           RAIN#lag_10             1.000000
    7           RAIN            RAIN#lag_9            -0.944911

    >>> top_n_correlations(res, goal_feature, n=2, grouped=False)
                      index  DEBIET_TOTAAL#lag_0
    0  DEBIET_TOTAAL#lag_10                 -1.0
    1       DEBIET_A#lag_10                  1.0
    """

    if goal_feature not in df.columns:
        raise ValueError("Goal feature not found in columns!")

    logging.debug(
        "Retrieving top n variables with goal variable {}, n={}, grouped={}".format(
            goal_feature, n, grouped
        )
    )

    pos_corr = df.corr().abs()  # get all positive correlations
    pos_corr = pos_corr.loc[goal_feature].reset_index()
    pos_corr = pos_corr.loc[pos_corr["index"] != goal_feature]

    if grouped:
        pos_corr["GROUP"] = pos_corr["index"].apply(lambda x: x.split(sep)[0])
        pos_corr = (
            pos_corr.groupby("GROUP")
            .apply(lambda x: x.nlargest(n, goal_feature))[["index", goal_feature]]
            .reset_index(drop=False)
        )
        pos_corr = pos_corr.drop("level_1", axis=1)

    else:
        pos_corr = pos_corr.sort_values(goal_feature, ascending=False).head(n)

    corrs = df.corr()  # replace correlations with the correct negative ones
    corrs = corrs.loc[goal_feature].reset_index()
    corrs = pos_corr.drop(goal_feature, axis=1).merge(corrs, on="index", how="left")
    logging.info("Output from retrieve_top_n_correlations:")
    log_dataframe_characteristics(corrs, logging.INFO)
    return corrs


def top_score_correlations(df: pd.DataFrame, goal_feature: str, score: float = 0.5):
    """
    Returns the features that have a correlation
    above a certain threshold with the defined goal feature

    Parameters
    ----------
    df: pd.DataFrame
        Dataframe containing the features that have to be correlated.

    goal_feature: str
        Feature that is used to compare the correlation with other features

    score: float (default: 0.5)
        absolute minimal correlation value

    Returns
    -------
    corrs: pd.DataFrame
        A dataframe containing 2 columns(index, goal feature).
        Index contains the correlating features and
        goal feature the correlation values.

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.feature_engineering import BuildRollingFeatures
    >>> from sam.exploration import top_score_correlations
    >>> import numpy as np
    >>> goal_feature = 'DEBIET_TOTAAL#lag_0'
    >>> df = pd.DataFrame({
    ...                'RAIN': [0.1, 0.2, 0.0, 0.6, 0.1, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ...                'DEBIET_A': [1, 2, 3, 4, 5, 5, 4, 3, 2, 4, 2, 3],
    ...                'DEBIET_B': [3, 1, 2, 3, 3, 6, 4, 1, 3, 3, 1, 5]})
    >>> df['DEBIET_TOTAAL'] = df['DEBIET_A'] + df['DEBIET_B']
    >>> RollingFeatures = BuildRollingFeatures(rolling_type='lag', \\
    ...     window_size = np.arange(10), lookback=0, keep_original=False)
    >>> res = RollingFeatures.fit_transform(df)
    >>> top_score_correlations(res, goal_feature, score=0.8)
                     index  DEBIET_TOTAAL#lag_0
    0  DEBIET_TOTAAL#lag_9             0.944911
    1           RAIN#lag_9            -0.944911
    2       DEBIET_B#lag_0             0.897340
    3           RAIN#lag_8             0.871695
    4       DEBIET_A#lag_0             0.838591
    """

    if goal_feature not in df.columns:
        raise ValueError("Goal feature not found in columns!")
    logging.debug(
        "Retrieving top n variables with goal variable {}, score={}".format(goal_feature, score)
    )

    pos_corr = df.corr().abs()  # get all positive correlations
    pos_corr = pos_corr.loc[goal_feature].reset_index()
    pos_corr = pos_corr.loc[
        (pos_corr["index"] != goal_feature) & (pos_corr[goal_feature] >= score)
    ]
    pos_corr = pos_corr.sort_values(goal_feature, ascending=False)

    corrs = df.corr()  # replace correlations with the correct negative ones
    corrs = corrs.loc[goal_feature].reset_index()
    corrs = pos_corr.drop(goal_feature, axis=1).merge(corrs, on="index", how="left")
    logging.info("Output from retrieve_top_score_correlations:")
    log_dataframe_characteristics(corrs, logging.INFO)
    return corrs



import numpy as np
import pandas as pd


def plot_lag_correlation(df: pd.DataFrame, ylim_min: int = None, ylim_max: int = None):
    """
    Visualize the correlations for rolling features

    Parameters
    ----------
    df: pandas dataframe (columns: LAG and columns to visualise)
        Dataframe containing the data that has to be visualized.

    ylim_min: int (default = None)
        minimun for customizing the y-axis limit.
    ylim_max: int (default = None)
        maximum for customizing the y-axis limit.

    Returns
    -------
    ax: seaborn plot
        Plot containing the visualization of the data

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.feature_engineering import BuildRollingFeatures
    >>> from sam.exploration import lag_correlation
    >>> from sam.visualization.rolling_correlations import plot_lag_correlation
    >>> import numpy as np
    >>> goal_feature = 'DEBIET#TOTAAL'
    >>> df = pd.DataFrame({
    ...                'RAIN': [0.1, 0.2, 0.0, 0.6, 0.1, 0.0,
    ...                         0.0, 0.0, 0.0, 0.0, 0.0, 0.0],
    ...                'DEBIET#A': [1, 2, 3, 4, 5, 5, 4, 3, 2, 4, 2, 3],
    ...                'DEBIET#B': [3, 1, 2, 3, 3, 6, 4, 1, 3, 3, 1, 5]})
    >>> df['DEBIET#TOTAAL'] = df['DEBIET#A'] + df['DEBIET#B']
    >>> test = lag_correlation(df, goal_feature)
    >>> f = plot_lag_correlation(test)
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    _, ax = plt.subplots(figsize=(15, 10))
    plt.axhline(0, ls="--", c="k")

    df_long = df.melt(id_vars="LAG")
    sns.lineplot(
        data=df_long,
        x="LAG",
        hue="variable",
        y="value",
        ax=ax,
        ls="--",
        marker="o",
        ms=12,
    )

    if ylim_min and ylim_max:
        ax.set_ylim(bottom=ylim_min, top=ylim_max)

    for col in df.drop("LAG", axis=1).columns:
        maxi = np.argmax(np.abs(df[col].values))
        plt.plot(df["LAG"].iloc[maxi], df[col].iloc[maxi], marker="o", ms=8, c="w")

    ax.set_ylabel("Correlation")
    ax.legend(loc="center right", bbox_to_anchor=(0.35, 0.5, 1, 0))
    sns.despine()
    plt.tight_layout()

    return ax



import logging

import pandas as pd

logger = logging.getLogger(__name__)


def log_new_columns(
    new: pd.DataFrame,
    old: pd.DataFrame,
    level: int = logging.INFO,
    log_no_changes: bool = False,
):
    """
    Given two pandas dataframe (new and old), checks what columns were removed/added, and logs them

    Logs what columns were removed (if there were any)
    Logs what columns were added (if there were any)
    Only logs column names. Doesn't check content of columns
    By default, logs if there were no changes, but this can be turned off.

    Parameters
    ----------
    new: pd.DataFrame
        The new pandas dataframe
    old: pd.DataFrame
        The old pandas dataframe
    level: loglevel (int) (default=logging.INFO)
        Loglevel to log at.
    log_no_changes : bool (default=False)
        if it should be logged when columns are identical.


    Returns
    -------
    Nothing

    Examples
    --------
    >>> from sam.logging_functions import log_new_columns
    >>> df = pd.DataFrame({'A': [1,2,3], 'C': [4,3,2]})
    >>> new_df = df.rename(columns = {'A': 'B'})
    >>> log_new_columns(new_df, df)
    """
    new_columns = list(set(new.columns.values) - set(old.columns.values))
    removed_columns = list(set(old.columns.values) - set(new.columns.values))

    if len(removed_columns) > 0:
        logger.log(level, "Removed columns: {}".format(removed_columns))
    if len(new_columns) > 0:
        logger.log(level, "Added new columns: {}".format(new_columns))
    if len(removed_columns) == 0 and len(new_columns) == 0 and log_no_changes:
        logger.log(level, "No changes in columns")



from typing import Sequence

import numpy as np
import pandas as pd


def tilted_loss(y_true: np.ndarray, y_pred: np.ndarray, quantile: float = 0.5, axis: int = -1):
    r"""
    Calculate tilted (also known as pinball or quantile loss) with numpy. Given a quantile
    :math:`q`, and an error :math:`e=y_\text{true}-y_\text{pred}`, then tilted loss :math:`L_q`
    is defined as:

    .. math::
        L_q(e) = \sum_{i:e_i\leq0} (1 - q) |e_i| + \sum_{i:e_i > 0} q |e_i|\\
        \text{where } i \in \{0, \dots, n_\text{samples} - 1\} \text{ and } q \in [0, 1]

    This function is equivalent to the mean absolute error if :math:`q=0.5`, which approximates
    the median:

    .. math::
       \begin{eqnarray}
       L_\text{0.5} &= \sum_{i:e_i\leq0} 0.5 |e_i| + \sum_{i:e_i > 0} 0.5 |e_i| \\
                    &= 0.5 * \sum_i |e_i|\\
       \text{MAE} &= \frac{1}{n_\text{samples}} \sum_i |e_i|
       \text{MAE} &= 2 L_\text{0.5}
       \end{eqnarray}

    For a given value of :math:`q`, the function that minimizes the tilted loss will be the
    :math:`q`'th quantile function.

    Parameters
    ----------
    y_true: array-like of shape = (n_samples, n_cols)
        True labels.
    y_pred: array-like of shape = (n_samples, n_cols)
        Predictions. Must be same shape as `y_true`
    quantile: float, optional (default=0.5)
        The quantile to use when computing tilted loss.
        Must be between 0 and 1 for tilted loss to be positive.
    axis: int, optional (default=-1)
        Over which axis to take the mean. By default the last axis (-1) is taken. For a
        (n_samples, n_cols) target array set axis=0 to retrieve the tilted loss of each col.

    Returns
    -------
    float:
        The quantile loss

    Examples
    --------
    >>> import numpy as np
    >>> from sam.metrics import tilted_loss
    >>> actual = np.array([1, 2, 3, 4])
    >>> pred = np.array([0.9, 2.1, 2.9, 3.1])
    >>> tilted_loss(actual, pred, quantile=0.5)
    0.15000000000000002
    """
    y_true, y_pred = np.array(y_true), np.array(y_pred)
    e = y_true - y_pred

    return np.mean(np.maximum(quantile * e, (quantile - 1) * e), axis=axis)


def joint_mae_tilted_loss(
    y_true: pd.DataFrame,
    y_pred: pd.DataFrame,
    quantiles: Sequence[float] = None,
    n_targets: int = 1,
):
    """
    Joint average and quantile regression loss function using mae.
    Sum of mean absolute error and multiple tilted loss functions
    Custom loss function, inspired by https://github.com/fmpr/DeepJMQR

    This calculates loss for multiple quantiles, and multiple targets.
    The total loss is the sum of the mae of all targets, and the tilted
    loss of all quantiles.

    ``y_true`` is expected to be a dataframe with shape ``(n_rows, n_targets)``
    ``y_pred`` is expected to be a dataframe with shape ``(n_rows, n_targets * (n_quantiles + 1))``
    For example, if there are 2 outputs and 2 quantiles, the order of y_pred
    should be: `[output_1_quantile_1, output_2_quantile_1,
    output_1_quantile_2, output_2_quantile_2, output_1_mean, output_2_mean]`

    Parameters
    ----------
    y_true: dataframe
        True values
    y_pred: dataframe
        Predicted values
    quantiles: sequence of floats (default=None)
        Quantiles to predict. Values between 0 and 1. By default, no quantile loss is used.
    n_targets: integer, optional (default=1)
        The number of distinct outputs to predict

    Returns
    -------
    float:
        The joint mae tilted loss

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from sam.metrics import joint_mae_tilted_loss
    >>> y_true = pd.DataFrame(np.array([4, 5, 6]), columns=["output_1"])
    >>> y_pred = pd.DataFrame(
    ...     np.array([[1, 2, 3], [7, 8, 9], [3.9, 5.1, 5.9]]),
    ...     columns=["output_1_quantile_1", "output_1_quantile_2", "output_1_mean"],
    ... )
    >>> joint_mae_tilted_loss(y_true, y_pred, quantiles=[0.1, 0.9], n_targets=1)
    3.44
    """
    if quantiles is None:
        quantiles = []

    average_prediction = y_pred.iloc[:, -1 * n_targets :].values
    # The last node will be fit with 2x the 0.5 quantile tilted loss (same as MAE)
    loss = 2 * np.sum(tilted_loss(y_true, average_prediction, quantile=0.5, axis=0))
    # For each quantile fit one node with corresponding tilted loss
    for i, q in enumerate(quantiles):
        # Select the i-th node
        q_pred = y_pred.iloc[:, n_targets * i : n_targets * (i + 1)].values
        # add tilted loss to total loss
        loss += np.sum(tilted_loss(y_true, q_pred, quantile=q, axis=0))
    return loss


def joint_mse_tilted_loss(
    y_true: pd.DataFrame,
    y_pred: pd.DataFrame,
    quantiles: Sequence[float] = None,
    n_targets: int = 1,
):
    """
    Joint average and quantile regression loss function using mse.
    Sum of mean absolute error and multiple tilted loss functions
    Custom loss function, inspired by https://github.com/fmpr/DeepJMQR

    This calculates loss for multiple quantiles, and multiple targets.
    The total loss is the sum of the mse of all targets, and the tilted
    loss of all quantiles.

    ``y_true`` is expected to be a dataframe with shape ``(n_rows, n_targets)``
    ``y_pred`` is expected to be a dataframe with shape ``(n_rows, n_targets * (n_quantiles + 1))``
    For example, if there are 2 outputs and 2 quantiles, the order of columns of y_pred
    should be: `[output_1_quantile_1, output_2_quantile_1,
    output_1_quantile_2, output_2_quantile_2, output_1_mean, output_2_mean]`

    Parameters
    ----------
    y_true: dataframe
        True values
    y_pred: dataframe
        Predicted values
    quantiles: sequence of floats (default=None)
        Quantiles to predict. Values between 0 and 1. By default, no quantile loss is used.
    n_targets: integer, optional (default=1)
        The number of distinct outputs to predict

    Returns
    -------
    float:
        The joint mse tilted loss

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from sam.metrics import joint_mae_tilted_loss
    >>> y_true = pd.DataFrame(np.array([4, 5, 6]), columns=["output_1"])
    >>> y_pred = pd.DataFrame(
    ...     np.array([[1, 2, 3], [7, 8, 9], [3.9, 5.1, 5.9]]),
    ...     columns=["output_1_quantile_1", "output_1_quantile_2", "output_1_mean"],
    ... )
    >>> joint_mse_tilted_loss(y_true, y_pred, quantiles=[0.1, 0.9], n_targets=1)
    7.410000000000002
    """
    if quantiles is None:
        quantiles = []

    average_prediction = y_pred.iloc[:, -1 * n_targets :].values
    # The last node will be fit with regular mean squared error
    loss = np.sum(np.mean((y_true - average_prediction) ** 2, axis=0))
    # For each quantile fit one node with corresponding tilted loss
    for i, q in enumerate(quantiles):
        # Select the i-th node
        q_pred = y_pred.iloc[:, n_targets * i : n_targets * (i + 1)].values
        # add tilted loss to total loss
        loss += np.sum(tilted_loss(y_true, q_pred, quantile=q, axis=0))
    return loss



import logging
from abc import ABC, abstractmethod
from typing import Callable, List

import numpy as np
import pandas as pd
from sklearn.base import TransformerMixin, BaseEstimator
from sklearn.utils.validation import check_is_fitted


class BaseFeatureEngineer(BaseEstimator, TransformerMixin, ABC):
    """
    Base class for feature engineering.
    To use this class, you need to implement the feature_engineer method.
    """

    def __init__(self):
        pass

    @abstractmethod
    def feature_engineer_(self, X) -> pd.DataFrame:
        """Implement this method to do the feature engineering."""
        raise NotImplementedError("You need to implement the feature_engineer_ method.")

    def fit(self, X, y=None):
        """fit method"""
        self._feature_names = self.feature_engineer_(X).columns.tolist()
        return self

    def transform(self, X) -> pd.DataFrame:
        """transform method"""
        logging.info("Feature engineering - input shape: %s", X.shape)
        X_out = self.feature_engineer_(X)
        logging.info("Feature engineering - output shape: %s", X_out.shape)
        return X_out

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Function for obtaining feature names. Generally used instead of the attribute, and more
        compatible with the sklearn API.

        Returns
        -------
        list:
            list of feature names
        """
        check_is_fitted(self, "_feature_names")
        return self._feature_names


class FeatureEngineer(BaseFeatureEngineer):
    """
    Feature engineering class. This class is used to feature engineer the data using default
    methods and makes integration with the timeseries models easier. You can implement your own
    feature engineering code as a function that takes two arguments: X and y and returns a
    feature table as a pandas dataframe.

    Parameters
    ----------
    feature_engineer_function : Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame]
        The feature engineering function.

    Example
    -------
    >>> from sam.feature_engineering.base_feature_engineering import FeatureEngineer
    >>> def feature_engineer(X, y=None):
    ...     X['C'] = X['A'] + X['B']
    ...     return X
    >>> df = pd.DataFrame({"A": [1, 2, 3, 4, 5], "B": [3, 4, 5, 6, 7]})
    >>> fe = FeatureEngineer(feature_engineer)
    >>> df_out = fe.fit_transform(df)
    """

    def __init__(
        self,
        feature_engineer_function: Callable[[pd.DataFrame, pd.DataFrame], pd.DataFrame] = None,
    ):
        self.feature_engineer_function = feature_engineer_function

    def feature_engineer_(self, X: pd.DataFrame) -> pd.DataFrame:
        """feature engineering function"""
        if self.feature_engineer_function is None:
            raise ValueError("You need to specify a feature engineering function.")
        return self.feature_engineer_function(X)


class IdentityFeatureEngineer(BaseFeatureEngineer):
    """
    Identity feature engineering class. This is a placeholder class for when you don't want to
    apply any feature engineering. Makes compatibility with the sam API easier.

    Parameters
    ----------
    numeric_only : bool
        Whether to only include numeric columns in the output.

    Example
    -------
    >>> from sam.feature_engineering.base_feature_engineering import IdentityFeatureEngineer
    >>> df = pd.DataFrame({"A": [1, 2, 3, 4, 5], "B": [3, 4, 5, 6, 7]})
    >>> fe = IdentityFeatureEngineer()
    >>> df_out = fe.fit_transform(df)
    """

    def __init__(self, numeric_only: bool = True):
        self.numeric_only = numeric_only

    def feature_engineer_(self, X: pd.DataFrame) -> pd.DataFrame:
        """feature engineering function, returns the input dataframe"""
        if self.numeric_only:
            return X.select_dtypes(include=np.number)
        return X



import logging
from abc import ABC, abstractmethod
from typing import List

import numpy as np
import pandas as pd
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

logger = logging.getLogger(__name__)


class BaseValidator(BaseEstimator, TransformerMixin, ABC):
    """Abstract base class for validators"""

    def __init__(self):
        pass

    @abstractmethod
    def validate(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Validate the data.

        This method should return a boolean array of the same shape as X, where True indicates a
        value that is invalid.
        """
        raise NotImplementedError("You need to implement the validate method.")

    def fit(self, X, y=None):
        """fit method"""
        self.feature_names_in_ = X.columns.tolist()
        self.feature_names_out_ = self.feature_names_in_
        return self

    def transform(self, X: pd.DataFrame) -> pd.DataFrame:
        """transform method"""
        X = X.copy()
        invalid_data = self.validate(X)
        X[invalid_data] = np.nan
        return X

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Function for obtaining feature names. Generally used instead of the attribute, and more
        compatible with the sklearn API.
        Returns
        -------
        list:
            list of feature names
        """
        check_is_fitted(self, "_feature_names_out_")
        return self._feature_names_out_



import warnings
from typing import Union

import numpy as np
import pandas as pd


def make_shifted_target(
    y: pd.Series,
    use_diff_of_y: bool = False,
    lags: Union[int, list] = 1,
    newcol_prefix: str = None,
):
    """
    Creates a target dataframe by performing shifting or differencing (once or multiple times)
    on a monospaced series. The dataframe contains columns 'TARGET_lead_x', where `x` are the
    lags and TARGET is the name of the input series.

    Parameters
    ----------
    y: pd.Series
        A series containing the target data. Must be monospaced in time, for the shifting
        to work correctly.
    use_diff_of_y: boolean, optional (default=False)
        If true, the difference between the target now and the shifted target is used,
        else just uses the shifted target itself.
    lags: array-like or int, optional (default=1)
        A list of integers, or a single integer describing what lags should be used to look in the
        future. For example, if this is `[1, 2, 3]`, the output will have three columns, performing
        shifting on 1, 2, and 3 timesteps in the future.
        If this is a list, the output will be a dataframe. If this is a scalar, the output will
        be a pd.Series
    newcol_prefix: str, optional (default=None)
        The prefix that the output columns will have. If None, `y.name` is used instead.

    Returns
    -------
    target: pd.DataFrame or pd.Series
        A target with the same index as `y`, and columns equal to `len(lags)`
        The values will be 'future values' of `y`.
        If we consider the index to be the 'timestamp', then the index will be the moment the
        prediction is made, not the moment the prediction is about. Therefore, the columns
        will be the future values with different lags.
        Any values that cannot be calculated (because there is no available future value) will be
        set to `np.nan`.

    Examples
    --------
    >>> df = pd.DataFrame({
    ...     'X': [18, 19, 20, 21],
    ...     'y': [10, 20, 50, 100]
    ... })
    >>> make_shifted_target(df['y'], use_diff_of_y = False, lags=1)
    0     20.0
    1     50.0
    2    100.0
    3      NaN
    Name: y_lead_1, dtype: float64

    --------
    >>> df = pd.DataFrame({
    ...     'X': [18, 19, 20, 21],
    ...     'y': [10, 20, 50, 100]
    ... })
    >>> make_shifted_target(df['y'], use_diff_of_y = True, lags=1)
    0    10.0
    1    30.0
    2    50.0
    3     NaN
    Name: y_diff_1, dtype: float64
    """

    if newcol_prefix is None:
        newcol_prefix = y.name

    if np.isscalar(lags):
        series_output = True
        lags = [lags]
    else:
        series_output = False

    for lag in lags:
        if lag < 0:
            raise ValueError("All lags must be greater or equal to 0")
        if lag % 1 != 0:
            raise ValueError("All lags must be integers")

    # Lagging to the future means negative lags
    if use_diff_of_y:
        result = pd.concat([-1 * y.diff(-1 * lag) for lag in lags], axis=1)
        names = ["{}_diff_{}".format(newcol_prefix, lag) for lag in lags]
    else:
        result = pd.concat([y.shift(-1 * lag) for lag in lags], axis=1)
        names = ["{}_lead_{}".format(newcol_prefix, lag) for lag in lags]

    result.columns = names
    if series_output:
        result = result.iloc[:, 0]

    return result


def make_differenced_target(
    y: pd.Series,
    lags: Union[int, list] = 1,
    newcol_prefix: str = None,
):
    """
    Creates a target dataframe by performing differencing (once or multiple times)
    on a monospaced series. The dataframe contains columns 'TARGET_lead_x', where `x` are the
    lags and TARGET is the name of the input series.

    Parameters
    ----------
    y: pd.Series
        A series containing the target data. Must be monospaced in time, for the differencing
        to work correctly.
    lags: array-like or int, optional (default=1)
        A list of integers, or a single integer describing what lags should be used to look in the
        future. For example, if this is `[1, 2, 3]`, the output will have three columns, performing
        differencing on 1, 2, and 3 timesteps in the future.
        If this is a list, the output will be a dataframe. If this is a scalar, the output will
        be a pd.Series
    newcol_prefix: str, optional (default=None)
        The prefix that the output columns will have. If None, `y.name` is used instead.

    Returns
    -------
    target: pd.DataFrame or pd.Series
        A target with the same index as `y`, and columns equal to `len(lags)`
        The values will be 'future values' of `y` but differenced.
        If we consider the index to be the 'timestamp', then the index will be the moment the
        prediction is made, not the moment the prediction is about. Therefore, the columns
        will be different future values with different lags.
        Any values that cannot be calculated (because there is no available future value) will be
        set to `np.nan`.

    Examples
    --------
    >>> df = pd.DataFrame({
    ...     'X': [18, 19, 20, 21],
    ...     'y': [10, 20, 50, 100]
    ... })
    >>> make_differenced_target(df['y'], lags=1)
    0    10.0
    1    30.0
    2    50.0
    3     NaN
    Name: y_diff_1, dtype: float64
    """
    warnings.warn(
        "make_differenced_target will be deprecated and replaced by make_shifted_target",
        PendingDeprecationWarning,
    )

    result = make_shifted_target(y=y, use_diff_of_y=True, lags=lags, newcol_prefix=newcol_prefix)

    return result


def inverse_differenced_target(predictions: pd.DataFrame, y: pd.Series):
    """
    Inverses differencing by adding the current values to the prediction.

    This function will take differenced target(s) and the current values, and return the actual
    target(s). Can be used to convert predictions from a differenced model to real predictions.

    predictions and y must be joined on index. Any indexes that only appear in predictions, or
    only appear in y, will also appear in the output, with nans inserted.

    Parameters
    ----------
    predictions: pd.DataFrame
        Dataframe containing differenced values.
    y: pd.Series
        The actual values in the present

    Returns
    -------
    actual: pd.DataFrame
        Dataframe containing un-differenced values, created by adding predictions to `y` on index.
        The index of this output will be the union of the indexes of predictions and `y`.
        The columns refer to the values/predictions made at a single point in time.
        For example, if the index is '18:00', and the predictions are made on differencing 1 hour,
        2 hour and 3 hours, then one row will contain the predictions made at 18:00,
        predicting what the target will be at 19:00, 20:00 and 21:00

    Examples
    --------
    >>> df = pd.DataFrame({
    ...     'X': [18, 19, 20, 21],
    ...     'y': [10, 20, 50, 100]
    ... })
    >>> target = make_differenced_target(df['y'], lags=1)
    >>> inverse_differenced_target(target, df['y'])
    0     20.0
    1     50.0
    2    100.0
    3      NaN
    Name: y_diff_1, dtype: float64

    >>> prediction = pd.DataFrame({
    ...    'pred_diff_1': [15, 25, 34, np.nan],
    ...    'pred_diff_2': [40, 55, np.nan, np.nan]
    ... })
    >>> inverse_differenced_target(prediction, df['y'])
       pred_diff_1  pred_diff_2
    0         25.0         50.0
    1         45.0         75.0
    2         84.0          NaN
    3          NaN          NaN
    >>> # This means that at timestep 0, we predict that the next two values will be 25 and 60
    >>> # At timestep 1, we predict the next two values will be 45 and 105
    >>> # At timestep 2, we predict the next two values will be 84 and unknown, etcetera.
    """
    result = predictions.add(y, axis=0)
    # Series lose their name on pd.Series.add, so put the name back
    if isinstance(result, pd.Series):
        result = result.rename(predictions.name)
    return result



import logging
from datetime import datetime
from typing import Callable, List, Union

import pandas as pd

logger = logging.getLogger(__name__)


def _validate_dataframe(df: pd.DataFrame):
    if df["ID"].isnull().sum() > 0:
        raise ValueError("ID column may not contain nans!")
    if df["TYPE"].isnull().sum() > 0:
        raise ValueError("TYPE column may not contain nans!")


def normalize_timestamps(
    df: pd.DataFrame,
    freq: str,
    start_time: Union[datetime, str] = "",
    end_time: Union[datetime, str] = "",
    round_method: str = "ceil",
    aggregate_method: Union[str, Callable, dict, List[Callable]] = "last",
    fillna_method: str = None,
):
    """
    Create a dataframe with all timestamps according to a given frequency. Fills in values
    for these timestamps from a given dataframe in SAM format.

    WARNING: This function makes assumptions about the data, and may not be safe for all datasets.
    For instance, a dataset with timestamps randomly distributed like a poisson process will
    significantly change when normalizing them, which may mean throwing away data.
    Furthermore, grouping measurements in the same timestamp may destroy cause-and-effect.
    For example, if a cause is measured at 15:58, and an effect is measured at 15:59, grouping
    them both at 16:00 makes it impossible to learn which came first.
    Use this function with caution, and generally mainly when the data already has normalized
    or close-to-normalized timestamps already.

    The process consists of four steps:

    Firstly, 'normalized' date ranges are created according to the required frequency.
    The start/end times of these date ranges can be given by start_time/end_time. If not given,
    the global minimum/maximum across all TYPE/ID is used. For example, if ID='foo' runs from 2017
    to 2019, and ID='bar' runs from 2018 to 2019, then ID='bar' will have missing values in
    the entirety of 2017.

    Secondly, all timestamps are rounded to the required frequency. For example, if the frequency
    is 1 hour, we may want the timestamp 19:45:12 to be rounded to 20:00:00.
    The method of rounding is ceiling by default, and is given by `round_method`.

    Thirdly, any timestamps with multiple measurements are aggregated.
    This is the last non-null value by default, and is given by `aggregate_method`.
    Other options are 'mean', 'median', 'first', and other pandas aggregation functions.

    Fourthly, any timestamps with missing values are filled. By default, no filling is done,
    and is given by `fillna_method`. The other options are backward filling and forward filling.
    ('bfill' and 'ffill')

    Parameters
    ----------
    df: pandas dataframe with TIME, TYPE, ID and VALUE columns, shape = (nrows, 4)
        Dataframe from which the values are created

    freq: str or DateOffset
        The required frequency for the time features.
        frequencies can have multiples, e.g. "15 min" for 15 minutes
        `See here for options
        <https://pandas.pydata.org/pandas-docs/stable/timeseries.html#offset-aliases>`_

    start_time: str or datetime-like, optional (default = '')
        the start time of the period to create features over
        if string, the format `%Y-%m-%d %H:%M:%S` will always work
        Pandas also accepts other formats, or a `datetime` object

    end_time: str or datetime-like, optional (default = '')
        the end time of the period to create features over
        if string, the format `%Y-%m-%d %H:%M:%S` will always work
        Pandas also accepts other formats, or a `datetime` object

    round_method: string, optional (default = 'floor')
        How to group the times in bins. By default, rows are grouped by
        flooring them to the frequency (e.g.: if frequency is hourly, the timestamp 18:59 will
        be grouped together with 18:01, and the TIME will be set to 18:00)
        The options are:

        - 'floor': Group times by flooring to the nearest frequency
        - 'ceil': Group times by ceiling to the nearest frequency
        - 'round': Group times by rounding to the nearest frequency

        Ceiling is the option that is the safest to prevent leakage: It will guarantee that a
        value in the output will have a `TIME` that is not before the time that it
        actually occurred.

    aggregate_method: function, string, dictionary, list of string/functions (default = 'last')
        Method that is used to aggregate values when multiple values fall
        within a specified frequency region.
        For example, when you have data per 5 minutes, but you're creating a
        an hourly frequency, the values need to be aggregated.
        Can be strings such as mean, sum, min, max, or a function.
        `See also
        <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.aggregate.html>`_

    fillna_method: string, optional (default = None)
        Method used to fill NA values, must be an option from `pd.DataFrame.fillna`.
        Options are: 'backfill', 'bfill', 'pad', 'ffill', None
        `See also
        <https://pandas.pydata.org/pandas-docs/stable/generated/pandas.DataFrame.fillna.html>`_

    Returns
    -------
    complete_df: pandas dataframe,
        shape `(length(TIME) * length(unique IDs) * length(unique TYPEs, 4))`
        dataframe containing all possible combinations of timestamps and IDs and TYPEs
        with selected frequency, aggregate method and fillna method

    Examples
    --------
    >>> from sam.preprocessing import normalize_timestamps
    >>> from datetime import datetime
    >>> import pandas as pd
    >>> df = pd.DataFrame({'TIME': [datetime(2018, 6, 9, 11, 13), datetime(2018, 6, 9, 11, 34),
    ...                             datetime(2018, 6, 9, 11, 44), datetime(2018, 6, 9, 11, 46)],
    ...                    'ID': "SENSOR",
    ...                    'TYPE': "DEPTH",
    ...                    'VALUE': [1, 20, 3, 20]})
    >>>
    >>> normalize_timestamps(df, freq = "15 min", end_time="2018-06-09 12:15:00")
                     TIME      ID   TYPE  VALUE
    0 2018-06-09 11:15:00  SENSOR  DEPTH    1.0
    1 2018-06-09 11:30:00  SENSOR  DEPTH    NaN
    2 2018-06-09 11:45:00  SENSOR  DEPTH    3.0
    3 2018-06-09 12:00:00  SENSOR  DEPTH   20.0
    4 2018-06-09 12:15:00  SENSOR  DEPTH    NaN

    >>> from sam.preprocessing import normalize_timestamps
    >>> from datetime import datetime
    >>> import pandas as pd
    >>> df = pd.DataFrame({'TIME': [datetime(2018, 6, 9, 11, 13), datetime(2018, 6, 9, 11, 34),
    ...                             datetime(2018, 6, 9, 11, 44), datetime(2018, 6, 9, 11, 46)],
    ...                    'ID': "SENSOR",
    ...                    'TYPE': "DEPTH",
    ...                    'VALUE': [1, 20, 3, 20]})
    >>>
    >>> normalize_timestamps(df, freq = "15 min", end_time="2018-06-09 12:15:00",
    ...                     aggregate_method = "mean", fillna_method="ffill")
                     TIME      ID   TYPE  VALUE
    0 2018-06-09 11:15:00  SENSOR  DEPTH    1.0
    1 2018-06-09 11:30:00  SENSOR  DEPTH    1.0
    2 2018-06-09 11:45:00  SENSOR  DEPTH   11.5
    3 2018-06-09 12:00:00  SENSOR  DEPTH   20.0
    4 2018-06-09 12:15:00  SENSOR  DEPTH   20.0
    """

    if df.empty:
        raise ValueError("No dataframe found")
    df = df.copy()

    _validate_dataframe(df)

    original_rows = df.shape[0]
    original_nas = df["VALUE"].isna().sum()
    timezone = df["TIME"].dt.tz

    fillna_options = ["backfill", "bfill", "pad", "ffill", None]
    if fillna_method not in fillna_options:
        raise ValueError("fillna_method not in {}".format(str(fillna_options)))

    round_method_options = ["floor", "round", "ceil"]
    if round_method == "floor":
        df["TIME"] = df["TIME"].dt.floor(freq)  # Technically, this has no effect on the result
    elif round_method == "round":
        df["TIME"] = df["TIME"].dt.round(freq)
    elif round_method in "ceil":
        df["TIME"] = df["TIME"].dt.ceil(freq)
    else:
        raise ValueError("round_method not in {}".format(str(round_method_options)))

    if not start_time:
        start_time = df["TIME"].min()
    if not end_time:
        end_time = df["TIME"].max()

    logger.debug(
        "Normalizing timestamps: freq={}, start_time={}, end_time={}, "
        "aggregate_method={}, fillna_method={}, round_times={}".format(
            freq, start_time, end_time, aggregate_method, fillna_method, round_method
        )
    )

    ids, types, time = pd.core.reshape.util.cartesian_product(
        [
            df["ID"].unique(),
            df["TYPE"].unique(),
            pd.date_range(start=start_time, end=end_time, freq=freq, tz=timezone),
        ]
    )

    complete_df = pd.DataFrame(dict(TIME=time, ID=ids, TYPE=types), columns=["TIME", "ID", "TYPE"])
    complete_df["TIME"] = complete_df["TIME"].dt.floor(freq)

    # Function currently groups based on first left matching frequency,
    # can be set to the first right frequency within the Grouper function
    df = df.groupby([pd.Grouper(key="TIME", freq=freq), "ID", "TYPE"]).agg(
        {"VALUE": aggregate_method}
    )

    df = df.reset_index(drop=False)

    complete_df = complete_df.merge(df, how="left", on=["TIME", "ID", "TYPE"])

    logger.debug("Number of missings before fillna: {}".format(complete_df["VALUE"].isna().sum()))

    if fillna_method:
        complete_df["VALUE"] = complete_df.groupby(["ID", "TYPE"])["VALUE"].apply(
            lambda x: x.fillna(method=fillna_method)
        )

    logger.info(
        "Dataframe changed because of normalize_timestamps: "
        "Previously it had {} rows, now it has {}".format(original_rows, complete_df.shape[0])
    )
    logger.info(
        "Also, the VALUE column previously had {} missing values, now it has {}".format(
            original_nas, complete_df["VALUE"].isna().sum()
        )
    )

    return complete_df



from typing import List, Union

import numpy as np
import pandas as pd


def compute_quantile_ratios(
    y: pd.Series, pred: pd.DataFrame, predict_ahead: int = 0, precision: int = 3
):
    """
    Computes the total proportion of data points in y beneath each quantile in pred.
    So for example, with quantile 0.75, you'd expect 75% of values in y to be beneath this quantile
    This function would then return a dictionary with 0.75 as key, and the actual proportion of
    values that fell beneath that quantile in y (e.g. 0.743).
    This allows the user to judge whether the fitted quantile model accurately reflects the
    distribution of the data.

    Parameters
    ---------
    y: pd.Series
        series of observed true values
    pred: pd.DataFrame
        output of MLPTimeseriesRegressor.predict() function
    predict_ahead: int (default=0)
        Number of timestep ahead to evaluate
    precision: int (default=3)
        decimal point precision of result

    Returns
    -------
    quantile_ratios: dict
        key is quantile, value is the observed proportion of data points below that quantile
    """

    # derive quantiles from column names
    qs = [float(c.split("_")[-1]) for c in pred.columns if "mean" not in c]

    quantile_ratios = {
        # mean here computes ratio (mean of True/False - 0/1s)
        q: (y < pred["predict_lead_%d_q_" % predict_ahead + str(q)]).mean()
        for q in qs
    }

    quantile_ratios = {
        round(key, precision): round(value, precision) for key, value in quantile_ratios.items()
    }

    return quantile_ratios


def compute_quantile_crossings(
    pred: pd.DataFrame, predict_ahead: int = 0, qs: List[Union[float, str]] = None
):
    """
    Computes the total proportion of predictions of a certain quantile that fall below the next
    lower quantile. This phenomenon is called 'quantile crossing'.
    So for example, with quantiles [0.1, 0.25, 0.75, 0.9], you'd expect all predictions of
    the 0.9 quantile to be above the 0.75 quantile predictions.
    In this example, this function would calculate the proportion of observations where the 0.9
    prediction actually falls below the 0.75 prediction, the 0.75 below the 0.25, and the 0.25
    below the 0.1 predictions.
    In addition, this function calculates in what proportion of cases the mean prediction falls
    outside of the closest quantile border below and above 0.5.
    NOTE: this function operates on the output of MLPTimeseriesRegressor.predict()
    NOTE-2: this function expects that either 0.5 or 'mean' is in the qs, or in the columns of pred
    when qs=None.

    Parameters
    ---------
    pred: pd.DataFrame
        Output of MLPTimeseriesRegressor.predict() function
    predict_ahead: int (default=0)
        Number of timestep ahead to evaluate
    qs: list (default=None)
        List of quantiles. If none, uses all quantiles in the pred columns.
        qs can be provided to compare crossings between a specific subset of quantiles.
        You can also add 'mean' to this list to add it to the comparison. It will then
        be compared to the nearest quantiles above and below 0.5.

    Returns
    -------
    crossings: dict
        Key is quantile, value is the proportion of quantile crossings.
    """

    # switch mean with 0.5 for ease in rest of function
    if qs is None:
        qs = [float(c.split("_")[-1]) for c in pred.columns if "mean" not in c] + [0.5]
        if qs.count(0.5) > 1:
            raise ValueError("0.5 and 'mean' cannot both be in qs")
    else:
        if 0.5 in qs and "mean" in qs:
            raise ValueError("0.5 and 'mean' cannot both be in qs")
        qs = [0.5 if q == "mean" else q for q in qs]

    # now replace the 'mean' part with 0.5 in the predictions
    pred.columns = [c.replace("mean", "q_0.5") for c in pred.columns]

    # make sure quantiles are sorted if they aren't already:
    qs = np.sort(qs)[::-1]

    # now compute the quantile crossings
    crossings = {}
    for c in range(len(qs) - 1):
        crossings[f"{qs[c]:.3f} < {qs[c+1]:.3f}"] = (
            pred[f"predict_lead_{predict_ahead}_q_{qs[c]}"]
            < pred[f"predict_lead_{predict_ahead}_q_{qs[c+1]}"]
        ).mean()

    # now replace 0.5 with mean again
    crossings = {key.replace("0.500", "mean"): value for key, value in crossings.items()}

    return crossings



from typing import Union
import numpy as np
import pandas as pd

from sam.validation import BaseValidator


class OutsideRangeValidator(BaseValidator):
    """
    Validator class method that removes data that is outside the provided range

    Parameters
    ----------
    cols: list (optional)
        Columns of input data to be checkout for being outside range. If None, all columns will be
        validated
    min_value: float, dict or "auto" (optional)
        Minimum value to check against. If None, no minimum will be checked. If "auto", the minimum
        value of the data will be used.
    max_value: float, dict or "auto" (optional)
        Maximum value to check against. If None, no maximum will be checked. If "auto", the maximum
        value of the data will be used.
    """

    def __init__(
        self,
        cols: list = None,
        min_value: Union[float, dict, str] = None,
        max_value: Union[float, dict, str] = None,
    ):
        self.cols = cols
        self.min_value = min_value
        self.max_value = max_value

    def fit(self, X, y=None):
        """
        Fit the transformer to the data.

        Parameters
        ----------
        X: pd.DataFrame
            Dataframe containing the features to be checked.
        y: pd.Series or pd.DataFrame (optional)
            Series or dataframe containing the target (ignored)
        """

        if self.cols is None:
            self.cols = X.columns

        if self.min_value == "auto":
            self.min_value_ = X[self.cols].min().to_dict()
        elif self.min_value is None:
            self.min_value_ = -np.inf
        elif isinstance(self.min_value, str):
            raise ValueError("min_value must be a float, dict or 'auto'")
        else:
            self.min_value_ = self.min_value

        if self.max_value == "auto":
            self.max_value_ = X[self.cols].max().to_dict()
        elif self.max_value is None:
            self.max_value_ = np.inf
        elif isinstance(self.max_value, str):
            raise ValueError("max_value must be a float, dict or 'auto'")
        else:
            self.max_value_ = self.max_value

        self._feature_names_out = X.columns

        return self

    def validate(self, X):
        """
        Transform the data.

        Parameters
        ----------
        X: pd.DataFrame
            Dataframe containing the features to be checked.
        """

        invalid_data = pd.DataFrame(
            data=np.zeros_like(X.values).astype(bool),
            index=X.index,
            columns=X.columns,
        )

        invalid_data[self.cols] = X[self.cols].gt(self.max_value_) | X[self.cols].lt(
            self.min_value_
        )

        return invalid_data



from typing import Callable

from sam.metrics import keras_joint_mae_tilted_loss, keras_joint_mse_tilted_loss


def create_keras_quantile_mlp(
    n_input: int,
    n_neurons: int,
    n_layers: int,
    n_target: int = 1,
    quantiles: list = None,
    dropout: float = 0.0,
    momentum: float = 1.0,
    hidden_activation: str = "relu",
    output_activation: str = "linear",
    lr: float = 0.001,
    average_type: str = "mean",
) -> Callable:
    """
    Creates a multilayer perceptron in keras.
    Optimizes the keras_joint_mse_tilted_loss to do multiple quantile and
    mean/median regression with a single model.

    Parameters
    ----------
    n_input: int
        Number of input nodes
    n_neurons: int
        Number of neurons hidden layer
    n_layers: int
        Number of hidden layers. 0 implies that the output is no additional layer
        between input and output.
    n_target: int, optional (default=1)
        Number of distinct outputs. Each will have their own mean and quantiles
        When fitting the model, this should be equal to the number of columns in y_train
    quantiles: list of floats (default=None)
        Quantiles to predict, values between 0 and 1,
        default is None, which returns a regular mlp (single output)
        for mean squared error regression
    dropout: float, optional (default=0.0)
        Rate parameter for dropout, value in (0,1).
        Default is 0.0, which means that no batch dropout is applied
    momentum: float, optional (default=1.0)
        Parameter for batch normalization, value in (0,1)
        default is 1.0, which means that no batch normalization is applied
        Smaller values means stronger batch normalization, see keras documentation.
        https://keras.io/layers/normalization/
    hidden_activation: str (default='relu')
        Activation function for hidden layers, for more explanation:
        https://keras.io/layers/core/
    output_activation: str (default='linear')
        Activation function for output layer, for more explanation:
        https://keras.io/layers/core/
    lr: float (default=0.001)
        Learning rate
    average_type: str (default='mean')
        Determines what to fit as the average: 'mean', or 'median'. The average is the last
        node in the output layer and does not reflect a quantile, but rather estimates the central
        tendency of the data. Setting to 'mean' results in fitting that node with MSE, and
        setting this to 'median' results in fitting that node with MAE (equal to 0.5 quantile).

    Returns
    --------
        keras model

    Examples
    --------
    >>> from sam.models import create_keras_quantile_mlp
    >>> from sam.datasets import load_rainbow_beach
    >>> data = load_rainbow_beach()
    >>> X, y = data, data["water_temperature"]
    >>> n_input = X.shape[1]
    >>> n_neurons = 64
    >>> n_layers = 3
    >>> quantiles = [0.1, 0.5, 0.9]
    >>> model = create_keras_quantile_mlp(n_input, n_neurons, n_layers, quantiles=quantiles)
    >>> model.fit(X, y, batch_size=16, epochs=20, verbose=0)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    """
    from tensorflow.keras.layers import (
        Activation,
        BatchNormalization,
        Dense,
        Dropout,
        Input,
    )
    from tensorflow.keras.models import Model
    from tensorflow.keras.optimizers import Adam

    if quantiles is None:
        quantiles = []
    if dropout is None:
        dropout = 0.0
    if momentum is None:
        momentum = 1.0
    if len(quantiles) == 0:
        mse_tilted = "mse"
    else:

        def mse_tilted(y, f):
            if average_type == "mean":
                loss = keras_joint_mse_tilted_loss(y, f, quantiles, n_target)
            elif average_type == "median":
                loss = keras_joint_mae_tilted_loss(y, f, quantiles, n_target)
            return loss

    n_out = n_target * (len(quantiles) + 1)  # one extra for mean regression
    input_layer = Input((n_input,))
    h = input_layer

    # repeat adding the same type of layer
    for _ in range(n_layers):
        h = Dense(n_neurons)(h)
        if momentum < 1:
            h = BatchNormalization(momentum=momentum)(h)
        h = Activation(hidden_activation)(h)
        h = Dropout(rate=dropout)(h)
    out = Dense(n_out, activation=output_activation)(h)

    model = Model(inputs=input_layer, outputs=out)
    model.compile(loss=mse_tilted, optimizer=Adam(learning_rate=lr))

    return model


def create_keras_quantile_rnn(
    input_shape: tuple,
    n_neurons: int = 64,
    n_layers: int = 2,
    quantiles: list = None,
    n_target: int = 1,
    layer_type: str = "GRU",
    dropout: float = 0.0,
    recurrent_dropout: str = "dropout",
    hidden_activation: str = "relu",
    output_activation: str = "linear",
    lr: float = 0.001,
) -> Callable:
    """
    Creates a simple RNN (LSTM or GRU) with keras.
    Optimizes the keras_joint_mse_tilted_loss to do multiple quantile and
    mean regression with a single model.

    Parameters
    ----------
    input_shape: tuple,
        A shape tuple (integers) of single input sample, (window, number of features)
        where window is the parameter used in the preprocessing.RecurrentReshaper class
    n_neurons: int (default=64)
        Number of neurons hidden layer
    n_layers: int (default=2)
        Number of hidden layers. 0 implies that the output is no additional layer
        between input and output.
    quantiles: list of floats (default=None)
        Quantiles to predict, values between 0 and 1,
        default is None, which returns a regular rnn (single output)
        for mean squared error regression
    n_target: int, optional (default=1)
        Number of distinct outputs. Each will have their own mean and quantiles
        When fitting the model, this should be equal to the number of columns in y_train
    layer_type: str (default='GRU')
        Type of recurrent layer
        Options: 'LSTM' (long short-term memory) or 'GRU' (gated recurrent unit)
    dropout: float, optional (default=0.0)
        Rate parameter for dropout, value in (0,1)
        default is 0.0, which means that no batch dropout is applied
    recurrent_dropout: float or str, optional (default='dropout')
        Rate parameter for dropout, value in (0,1)
        default is 'dropout', which means that recurrent dropout is equal to
        dropout parameter (dropout between layers)
    hidden_activation: str (default='relu')
        Activation function for hidden layers, for more explanation:
        https://keras.io/layers/core/
    output_activation: str (default='linear')
        Activation function for output layer, for more explanation:
        https://keras.io/layers/core/
    lr: float (default=0.001)
        Learning rate

    Returns
        keras model

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> from sam.data_sources import synthetic_date_range, synthetic_timeseries
    >>> from sam.preprocessing import RecurrentReshaper
    >>> from sam.models import create_keras_quantile_rnn
    >>> dates = pd.Series(synthetic_date_range().to_pydatetime())
    >>> y = synthetic_timeseries(dates, daily = 2, noise = {'normal': 0.25}, seed=2)
    >>> y = y[~np.isnan(y)]
    >>> X = pd.DataFrame(y)
    >>> X_3d = RecurrentReshaper(window=24, lookback = 1).fit_transform(X)
    >>> X_3d = X_3d[24:]
    >>> y = y[24:]
    >>> input_shape = X_3d.shape[1:]
    >>> model = create_keras_quantile_rnn(input_shape, quantiles=[0.01, 0.99])
    >>> model.fit(X_3d, y, batch_size=32, epochs=5, verbose=0)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    """
    from tensorflow.keras.layers import GRU, LSTM, Dense, Input
    from tensorflow.keras.models import Model
    from tensorflow.keras.optimizers import Adam

    if quantiles is None:
        quantiles = []
    if dropout is None:
        dropout = 0.0
    if len(quantiles) == 0:
        mse_tilted = "mse"
    else:

        def mse_tilted(y, f):
            loss = keras_joint_mse_tilted_loss(y, f, quantiles, n_target)
            return loss

    # Recurrent layer is either LSTM or GRU
    if layer_type.upper() == "LSTM":
        RNN = LSTM
    elif layer_type.upper() == "GRU":
        RNN = GRU
    else:
        raise ValueError('Invalid layer_type, choose "LSTM" or "GRU"')
    if recurrent_dropout == "dropout":
        recurrent_dropout = dropout
    n_out = n_target * (len(quantiles) + 1)  # one extra for mean regression

    input_layer = Input(input_shape)
    h = input_layer
    for i in range(n_layers):
        # last layer returns last value of sequence:
        return_sequences = i < n_layers - 1
        h = RNN(
            n_neurons,
            activation=hidden_activation,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            return_sequences=return_sequences,
        )(h)
    out = Dense(n_out, activation=output_activation)(h)
    model = Model(inputs=input_layer, outputs=out)
    model.compile(loss=mse_tilted, optimizer=Adam(lr=lr))
    return model


def create_keras_autoencoder_mlp(
    n_input: int,
    encoder_neurons: list = [64, 16],
    dropout: float = 0.0,
    momentum: float = 1.0,
    hidden_activation: str = "relu",
    output_activation: str = "linear",
    lr: float = 0.001,
) -> Callable:
    """
    Function to create an MLP auto-encoder in keras
    Optimizes the mean squared error to reconstruct input,
    after passing input through bottleneck neural network.

    Parameters
    ----------
    n_input: int
        Number of input nodes
    encoder_neurons: list (default=[64, 16])
        List of integers, each representing the number of neurons per layer
        within the encoder. Decoder is reversed version of encoder.
        Last element is number of neurons in "representation" layer
        Example:
        If n_layers=[64, 12], number of features is 120,
        the number of neurons per layer is [120, 64, 12, 64, 120].
    dropout: float, optional (default=0.0)
        Rate parameter for dropout, value in (0,1)
        default is 0.0, which means that no batch dropout is applied
    momentum: float, optional (default=1.0)
        Parameter for batch normalization, value in (0,1)
        default is 1.0, which means that no batch normalization is applied
        Smaller values means stronger batch normalization, see keras documentation.
        https://keras.io/layers/normalization/
    hidden_activation: str (default='relu')
        Activation function for hidden layers, for more explanation:
        https://keras.io/layers/core/
    output_activation: str (default='linear')
        Activation function for output layers, for more explanation:
        https://keras.io/layers/core/
    lr: float (default=0.001)
        Learning rate

    Returns
        keras model

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.data_sources import synthetic_date_range, synthetic_timeseries
    >>> from sam.preprocessing import RecurrentReshaper
    >>> from sam.models import create_keras_autoencoder_mlp
    >>> dates = pd.Series(synthetic_date_range().to_pydatetime())
    >>> X = [synthetic_timeseries(dates, daily=2, noise={'normal': 0.25}, seed=i)
    ...      for i in range(100)]
    >>> X = pd.DataFrame(X)
    >>> model = create_keras_autoencoder_mlp(n_input=100)
    >>> model.fit(X.T, X.T, batch_size=32, epochs=5, verbose=0)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    """
    from tensorflow.keras.layers import (
        Activation,
        BatchNormalization,
        Dense,
        Dropout,
        Input,
    )
    from tensorflow.keras.models import Model
    from tensorflow.keras.optimizers import Adam

    if dropout is None:
        dropout = 0.0
    if momentum is None:
        momentum = 1.0

    input_layer = Input((n_input,))
    # encoder
    h = input_layer
    # For each n_neuron value, add a dense layer to the model
    n_neurons = encoder_neurons + encoder_neurons[:-1][::-1]
    for n in n_neurons:
        h = Dense(n)(h)
        if momentum < 1:
            h = BatchNormalization(momentum=momentum)(h)
        h = Activation(hidden_activation)(h)
        h = Dropout(rate=dropout)(h)
    # output layer
    out = Dense(n_input, activation=output_activation)(h)
    # compile
    model = Model(inputs=input_layer, outputs=out)
    model.compile(loss="mse", optimizer=Adam(lr=lr))
    return model


def create_keras_autoencoder_rnn(
    input_shape: tuple,
    encoder_neurons: list = [64, 16],
    layer_type: str = "GRU",
    dropout: float = 0.0,
    recurrent_dropout: float = 0.0,
    hidden_activation: str = "relu",
    output_activation: str = "linear",
    lr: float = 0.001,
) -> Callable:
    """
    Function to create a recurrent auto-encoder in keras
    Optimizes the mean squared error to reconstruct input,
    after passing input through bottleneck neural network.
    Reference:
    https://towardsdatascience.com/lstm-autoencoder-for-anomaly-detection-e1f4f2ee7ccf
    https://blog.keras.io/building-autoencoders-in-keras.html

    Parameters
    ----------
    input_shape: tuple,
        shape of single input sample, (recurrent steps, number of features)
    encoder_neurons: list (default=[64, 16])
        List of integers, each representing the number of neurons per layer
        within the encoder. Decoder is reversed version of encoder.
        Last element is number of neurons in "representation" layer
        Example:
        If n_layers=[64, 12], number of features is 120,
        the number of neurons per layer is [120, 64, 12, 64, 120].
    layer_type: str (default='GRU')
        Type of recurrent layer
        Options: 'LSTM' (long short-term memory) or 'GRU' (gated recurrent unit)
    dropout: float, optional (default=0.0)
        Rate parameter for dropout, value in (0,1)
        default is 0.0, which means that no batch dropout is applied
    recurrent_dropout: float, optional (default=0.0)
        Rate parameter for dropout, value in (0,1)
        default is 0.0, which means that recurrent dropout is equal to
        dropout parameter (dropout between layers)
    hidden_activation: str (default='relu')
        Activation function for hidden layers, for more explanation:
        https://keras.io/layers/core/
    output_activation: str (default='linear')
        Activation function for output layers, for more explanation:
        https://keras.io/layers/core/
    lr: float (default=0.001)
        Learning rate

    Returns
        keras model

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.data_sources import synthetic_date_range, synthetic_timeseries
    >>> from sam.preprocessing import RecurrentReshaper
    >>> from sam.models import create_keras_autoencoder_rnn
    >>> dates = pd.Series(synthetic_date_range().to_pydatetime())
    >>> y = synthetic_timeseries(dates, daily=2, noise={'normal': 0.25}, seed=2)
    >>> X = pd.DataFrame(y)
    >>> X_3d = RecurrentReshaper(window=24, lookback=1).fit_transform(X)
    >>> X_3d = X_3d[24:]
    >>> input_shape = X_3d.shape[1:]
    >>> model = create_keras_autoencoder_rnn(input_shape)
    >>> model.fit(X_3d, X_3d, batch_size=32, epochs=5, verbose=0)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    """
    from tensorflow.keras.layers import (
        GRU,
        LSTM,
        Dense,
        Input,
        RepeatVector,
        TimeDistributed,
    )
    from tensorflow.keras.models import Model
    from tensorflow.keras.optimizers import Adam

    if dropout is None:
        dropout = 0.0
    lookback = input_shape[0]
    n_features = input_shape[1]
    if layer_type.upper() == "LSTM":
        RNN = LSTM
    elif layer_type.upper() == "GRU":
        RNN = GRU
    else:
        raise ValueError('layer_type should be "LSTM" or "GRU"')

    input_layer = Input(input_shape)
    # encoder
    h = input_layer
    # For each n_neuron value, add an rnn layer to the model
    n_layers = len(encoder_neurons)
    for i in range(n_layers - 1):
        h = RNN(
            encoder_neurons[i],
            activation=hidden_activation,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            return_sequences=True,
        )(h)
    # Last layer of encoder: only return last value of the LSTM output sequence
    h = RNN(
        encoder_neurons[-1],
        activation=hidden_activation,
        dropout=dropout,
        recurrent_dropout=recurrent_dropout,
        return_sequences=False,
    )(h)
    # decoder
    # repeat 1d representation to feed to an rnn layer
    h = RepeatVector(lookback)(h)
    # Reverse layers to create symmetric model
    for i in range(n_layers - 1, 0, -1):
        h = RNN(
            encoder_neurons[i],
            activation=hidden_activation,
            dropout=dropout,
            recurrent_dropout=recurrent_dropout,
            return_sequences=True,
        )(h)
    out = TimeDistributed(Dense(n_features, activation=output_activation))(h)
    # compile
    model = Model(inputs=input_layer, outputs=out)
    model.compile(loss="mse", optimizer=Adam(lr=lr))
    return model



import logging

import numpy as np
import pandas as pd
from sam.feature_engineering import BuildRollingFeatures
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted

logger = logging.getLogger(__name__)


class RecurrentReshaper(BaseEstimator, TransformerMixin):
    """Reshapes a two-dimensional feature table into a three dimensional table
    sliding window table, usable for recurrent neural networks

    An important note is that this transformer assumes that the data is sorted by time already!
    So if the input dataframe is not sorted by time (in ascending order), the results will be
    completely wrong.

    Given an input array with shape `(n_samples, n_features)`, output array is of shape
    `(n_samples, lookback, n_features)`

    Parameters
    ----------
    window : integer
        Number of rows to look back
    lookback : integer (default=0)
        the features that are built will be shifted by this value.
        If target is in `X`, `lookback` should be greater than 0 to avoid leakage.
    remove_leading_nan : boolean
        Whether leading nans should be removed.
        Leading nans arise because there is no history for first samples

    Examples
    --------
    >>> from sam.data_sources import read_knmi
    >>> from sam.preprocessing import RecurrentReshaper
    >>> X = read_knmi('2018-01-01 00:00:00', '2018-01-08 00:00:00').set_index('TIME')
    >>> reshaper = RecurrentReshaper(window=7)
    >>> X3D = reshaper.fit_transform(X)
    """

    def _validate_params(self):
        """apply various checks to the inputs of the __init__ function
        throw value error or type error based on the result
        """
        if self.lookback < 0:
            raise ValueError("lookback cannot be negative!")
        if not np.isscalar(self.window):
            raise ValueError("window should be a scalar")

    def __init__(self, window, lookback=1, remove_leading_nan=False):
        self.window = window
        self.lookback = lookback
        self.remove_leading_nan = remove_leading_nan
        self.window_range = range(window - 1, -1, -1)
        self.start = window + lookback - 1
        logger.debug(
            "Initialized reshaping generator. window={}, lookback={}, "
            "remove_leading_nan={}".format(window, lookback, remove_leading_nan)
        )
        if self.remove_leading_nan:
            logger.warning(
                "remove_leading_nan is True and can lead to "
                "X and y of unequal size! Be sure that this is what you want."
            )

    def fit(self, X=None, y=None):
        """Calculates n_features

        Parameters
        ----------
        X : Two dimensional dfeature table
        y : optional, is ignored
        """
        self._validate_params()
        self.n_features_ = X.shape[1]
        self.lag_transformer_ = BuildRollingFeatures(
            rolling_type="lag",
            window_size=self.window_range,
            lookback=self.lookback,
            keep_original=False,
        )
        self.lag_transformer_.fit(X)
        return self

    def transform(self, X):
        """Transforms feature table X to apply rolling function and reshaping

        Parameters
        ----------
        X : pandas Dataframe or numpy array, shape = `(n_rows, n_features)`
            feature table (so no ID, TYPE, TIME columns) to transform to three dimensional

        Returns
        -------
        X_new : numpy array
            A three dimensional numpy array, moving windows over the features table `X`
        """
        check_is_fitted(self, "n_features_")
        # X needs to be pandas dataframe to use BuildRollingFeatures
        X = pd.DataFrame(X)
        n_samples = X.shape[0]
        X_lags = self.lag_transformer_.transform(X)
        X_new = np.reshape(X_lags.values, (n_samples, self.window, self.n_features_))
        # remove first elements that have no full history
        if self.remove_leading_nan:
            X_new = X_new[self.start :, :, :]
        return X_new



import warnings


def parametrized(dec):
    """Decorator to make a decorator parametrized"""

    def layer(*args, **kwargs):
        def repl(f):
            return dec(f, *args, **kwargs)

        return repl

    return layer


@parametrized
def add_future_warning(func, msg):
    """Decorator to add a FutureWarning to a function."""

    def wrapper(*args, **kwargs):
        warnings.warn(msg, FutureWarning)
        return func(*args, **kwargs)

    return wrapper



from typing import Iterable

import pandas as pd


def plot_feature_importances(importances: pd.DataFrame, feature_names: Iterable = None):
    """
    Create bar graph of feature importances, with highest first.
    Also creates aggregated features over lag features. For this, pass a list of features as
    feature_names. It accepts the output of MLPTimeseriesRegressor.quantile_feature_importances().
    Alternatively, you can format your own feature importances as a pandas DataFrame with columns
    as features and rows as potentially multiple random iterations.

    Parameters
    ----------
    importances: pd.DataFrame
        Dataframe with features as columns and potentially multiple random iterations as rows.
    feature_names: iterable of strings or None (default=None)
        Iterable of column names (or starting column names) to aggregate for. Every element of
        feature_names is the common start name of each feature, i.e.: importances for
        feature_1#lag_1 and feature_1#lag3 are summed.

    Returns
    -------
    fig: matplotlib.pyplot.Figure
        Bar plot of all features in importances. Error bars indicate variance over iterations
    fig_sum: matplotlib.pyplot.Figure
        Bar plot with feature importances summed over lag features.
        Error bars indicate variance over iterations.

    Examples
    --------
    >>> # One way to get to feature importances is to first fit a MLPTimeseriesRegressor.
    >>> # In this example, we assumed you did and refer to it as `model`.
    >>> from sam.visualization import plot_feature_importances
    >>> # note that we need a negative here, as default score function is a loss
    >>> # Example with a fictional dataset with only 2 features
    >>> import pandas as pd
    >>> import seaborn
    >>> from sam.models import MLPTimeseriesRegressor
    >>> from sam.feature_engineering import SimpleFeatureEngineer
    >>> from sam.datasets import load_rainbow_beach
    ...
    >>> data = load_rainbow_beach()
    >>> X, y = data, data["water_temperature"]
    >>> test_size = int(X.shape[0] * 0.33)
    >>> train_size = X.shape[0] - test_size
    >>> X_train, y_train = X.iloc[:train_size, :], y[:train_size]
    >>> X_test, y_test = X.iloc[train_size:, :], y[train_size:]
    ...
    >>> simple_features = SimpleFeatureEngineer(
    ...     rolling_features=[
    ...         ("wave_height", "mean", 24),
    ...         ("wave_height", "mean", 12),
    ...     ],
    ...     time_features=[
    ...         ("hour_of_day", "cyclical"),
    ...     ],
    ...     keep_original=False,
    ... )
    ...
    >>> model = MLPTimeseriesRegressor(
    ...     predict_ahead=(0,),
    ...     feature_engineer=simple_features,
    ...     verbose=0,
    ... )
    ...
    >>> model.fit(X_train, y_train)  # doctest: +ELLIPSIS
    <keras.callbacks.History ...
    >>> score_decreases = model.quantile_feature_importances(
    ...     X_test[:100], y_test[:100], n_iter=3, random_state=42)
    >>> importances = -model.quantile_feature_importances(X, y, random_state=42)
    >>> fig, fig_sum = plot_feature_importances(importances,
    ...     list(model.get_input_cols()) + model.get_feature_names_out())
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    def _create_plot(importances):
        f = plt.figure(figsize=(10, 3 + importances.shape[1] * 0.2))
        order = list(importances.mean(axis=0).sort_values(ascending=False).index)
        sns.barplot(data=importances, order=order, orient="h")
        sns.despine()
        plt.tight_layout()
        return f

    fig = _create_plot(importances)

    fig_sum = plt.figure()
    if feature_names is not None:
        # and now summed over lag features
        importances_sum = {}
        for feature in feature_names:
            if feature != "TIME":
                these_cols = [c for c in importances.columns if c.startswith(feature)]
                importances_sum[feature] = importances[these_cols].sum(axis=1)
        importances_sum = pd.DataFrame(importances_sum)

        fig_sum = _create_plot(importances_sum)

    return fig, fig_sum



import pandas as pd


def remove_target_nan(X, y, use_x=False):
    """
    Remove rows with nan that can't be used for fitting ML models

    Parameters
    ----------
    X: pd.DataFrame
        The independent variables used to 'train' the model
    y: pd.Series or pd.DataFrame
        Target data (dependent variable) used to 'train' the model.
    use_x: bool
        If True, remove rows with nan in X and y. Otherwise, remove rows with nan in y.

    """
    X, y = X.copy(), y.copy()
    targetnanrows = pd.DataFrame(y).isna().any(axis=1)
    if use_x:
        targetnanrows = targetnanrows | pd.DataFrame(X).isna().any(axis=1)
    X = X.loc[~targetnanrows]
    y = y.loc[~targetnanrows]

    return X, y


def remove_until_first_value(X, y):
    """
    Remove rows until the first value is available.

    Parameters
    ----------
    X: pd.DataFrame
        The independent variables used to 'train' the model
    y: pd.Series or pd.DataFrame
        Target data (dependent variable) used to 'train' the model.

    """
    X, y = X.copy(), y.copy()
    first_complete_index = X.dropna(axis=0, how="any").index[0]
    X = X.loc[first_complete_index:]
    y = y.loc[first_complete_index:]

    return X, y



from .base_validator import BaseValidator
from .outside_range_validator import OutsideRangeValidator
from .mad_validator import RemoveExtremeValues, MADValidator
from .flatline_validator import RemoveFlatlines, FlatlineValidator
from .setup_validation_pipeline import create_validation_pipe

__all__ = [
    "BaseValidator",
    "OutsideRangeValidator",
    "RemoveExtremeValues",
    "MADValidator",
    "RemoveFlatlines",
    "FlatlineValidator",
    "create_validation_pipe",
]



from math import ceil, floor
from typing import Iterable, Tuple

import pandas as pd


def plot_incident_heatmap(
    df: pd.DataFrame,
    resolution: str = "row",
    row_column: str = "id",
    value_column: str = "incident",
    time_column: str = None,
    normalize: bool = False,
    figsize: Iterable[Tuple[int, int]] = (24, 4),
    xlabel_rotation: int = 30,
    datefmt: str = None,
    **kwargs,
):
    """
    Create and return a heatmap for incident occurence. This can be used to visualize e.g.
    the count of outliers/threshold surpassings/warnings given over time.
    This returns a subplot object that can be shown or edited further.

    Parameters
    ----------
    df: Pandas DataFrame
        Contains the data to be plotted in long format
    resolution: string (default="row")
        The aggregation level to be plotted. Is either "row" when every
        row needs to be plotted, or e.g. H, D, W, M for aggregations over time.
    row_column: string (default="id")
        The column that used to split the rows of the heatmap
    value_column: string (default="incident")
        The column containing the count to be plotted in the heatmap
    time_column: string (default=None)
        The column containing the time for the x-axis. When left None, the
        index will be used. Aggregation based on the resolution parameter
        is done on this column.
    normalize: boolean (default=False)
        Normalize the aggregated values
    figsize: tuple of floats (default=(24,4))
        Size of the output figure, must be set before initialization.
    xlabel_rotation: numeric, optional (default=30)
        The rotation of the x-axis date labels. Rotation is counterclockwise, beginning
        with the text lying horizontally. By default, rotate 30 degrees.
    datefmt: string, optional (default=None)
        Optionally, the format of the x-axis date labels. By default, use
        `%Y-%m-%dT%H:%M:%S%f`
    **kwargs:
        any arguments to pass to sns.heatmap()

    Returns
    -------
    plot:  matplotlib.axes._subplots.AxesSubplot object
        a plot containing the heatmap. Can be edited further,
        or printed to the output.

    Examples
    --------
    >>> from sam.visualization import plot_incident_heatmap
    >>> import pandas as pd
    >>> import numpy as np
    >>> # Initialize a random dataframe
    >>> rng = pd.date_range('1/1/2011', periods=150, freq='D')
    >>> ts = pd.DataFrame({'values': np.random.randn(len(rng)),
    ...                    'id': np.random.choice(['A','B','C'], len(rng))},
    ...                     index=rng, columns=['values','id'])
    >>>
    >>> # Create some incidents
    >>> ts['incident'] = 0
    >>> ts.loc[ts['values'] > .5, 'incident'] = 1
    >>>
    >>> # Create the heatmap
    >>> fig = plot_incident_heatmap(
    ...     ts, resolution='W', annot=True, cmap='Reds', datefmt="%Y, week %W"
    ... )
    """
    import matplotlib.pyplot as plt
    import seaborn as sns

    df = df.copy()

    # Resample the data if needed
    if resolution != "row":
        df_grouped = df.groupby([row_column, pd.Grouper(key=time_column, freq=resolution)])
        df_grouped = df_grouped[value_column].sum().unstack(fill_value=0)
    else:
        if time_column is not None:
            # Use a specific column instead of the index
            df_grouped = df.pivot(row_column, time_column, value_column).fillna(0)
        else:
            # Use the index, so first reset_index
            df_grouped = df.reset_index().pivot(row_column, "index", value_column).fillna(0)

    if normalize:
        df_grouped = df_grouped / df_grouped.values.max()

    # We typically want a wide figure
    plt.figure(figsize=figsize)

    # Initialize heatmap
    ax = sns.heatmap(df_grouped, **kwargs)
    ax.set_xticklabels(ax.get_xticklabels(), rotation=xlabel_rotation)

    # Set y limits. On some versions/platforms of seaborn, these are set at 0.5
    # instead of 0, so we round up the higher limit, and round down the lower limit.
    # If the limits are correctly set as integers, these lines have no effect.
    # It is totally possible for ax.get_ylim() to return (high, low), in which
    # case we need to switch them with `sorted` to obtain (low, high) instead.
    y_low, y_high = sorted(ax.get_ylim())
    ax.set_ylim(floor(y_low), ceil(y_high))

    # If desired, add custom date format to the x-axis
    if datefmt:
        ax.set_xticklabels(df_grouped.columns.strftime(datefmt))

    return ax



from .dataframe_functions import (
    assert_contains_nans,
    contains_nans,
    has_strictly_increasing_index,
    make_df_monotonic,
    sum_grouped_columns,
)
from .sklearnhelpers import FunctionTransformerWithNames
from .warnings import add_future_warning, parametrized

__all__ = [
    "assert_contains_nans",
    "contains_nans",
    "has_strictly_increasing_index",
    "make_df_monotonic",
    "sum_grouped_columns",
    "FunctionTransformerWithNames",
    "add_future_warning",
    "parametrized",
]



import logging
from typing import Union

import numpy as np
import pandas as pd
from sam.utils import add_future_warning
from sam.validation import BaseValidator

logger = logging.getLogger(__name__)


class FlatlineValidator(BaseValidator):
    """
    Detect flatlines and set to nan. Note that you have to check whether
    signals can contain natural flatliners (such as machines turned off),
    that might not need to be removed.

    Parameters
    ----------
    cols: list of strings (defaults to None)
        columns to apply this method to. If None, will apply to every column.
    window: "auto" or int (default = 1)
        number of previous equal values to consider current value a flatliner.
        so if set to 2, requires that 2 previous values are identical to
        current to set current value to nan.
        If set to "auto", the threshold is derived in the ``fit`` method.
        Based on a train set, the probability of difference being 0
        is estimated. This probability can be used to estimate the
        number of consecutive flatline samples, before the likelihood
        is below the ``pvalue`` parameter
        The maximum acceptable flatline window is derived for each column
        separately, with the same ``pvalue``
    pvalue: float or None (default=None)
        Threshold for likelihood of multiple consecutive flatline samples
        Only used if ``window="auto"``
        Small pvalues lead to a larger threshold, hence less flatlines will be
        removed
    margin: int (default = 0)
        Maximum absolute difference within window to consider them equal.
        Default is 0, which means that all samples within used window must be
        exactly equal to form a flatline.
    backfill: bool (default = True)
        whether to label all within the window, even before the first detected
        data point. This is useful if you want to remove flatlines from the
        beginning of a signal. However, that is not always representative of
        for a real-time application, so one might want to set this to False.

    Examples
    --------
    >>> import pandas as pd
    >>> from sam.validation import FlatlineValidator
    >>> # create some data
    >>> data = [1, 2, 6, 3, 4, 4, 4, 3, 6, 7, 7, 2, 2]
    >>> # with one clear outlier
    >>> test_df = pd.DataFrame()
    >>> test_df['values'] = data
    >>> # now detect flatlines
    >>> cols_to_check = ['values']
    >>> RF = FlatlineValidator(
    ...     cols=cols_to_check,
    ...     window=3)
    >>> data_corrected = RF.fit_transform(test_df)
    """

    def __init__(
        self,
        cols: list = None,
        window: Union[int, str] = 1,
        pvalue: float = None,
        margin: float = 0,
        backfill: bool = True,
    ):
        super().__init__()
        self.cols = cols
        self.window = window
        self.pvalue = pvalue
        self.margin = margin
        self.backfill = backfill

    def fit(self, data: pd.DataFrame):
        """If window size is 'auto', derive thresholds for each column
        Threshold is based on the probability that a sensor value does not change.
        The likelihood of a flatliner of m time steps, is this probability to the power m.
        A threshold such that flatliners with a likelihood below the pvalue are removed.

        Parameters
        ----------
        data: pd.DataFrame
            The dataframe to apply the method to
        """
        self.window_dict = {}
        for col in data.columns:
            if self.window == "auto":
                # estimate p
                p_estimate = (data[col].diff() == 0).mean()
                # compute threshold
                threshold = int(np.ceil(np.log(self.pvalue) / np.log(p_estimate)))
            else:
                threshold = self.window
            self.window_dict[col] = threshold
        return self

    def _validate_column(
        self,
        data: pd.Series,
        window: Union[int, str],
    ) -> pd.Series:
        """
        Validates a single column against the fitted dataframe
        """
        data = data.copy()

        # check if values within range are within margin
        flatliners = (
            ((data.rolling(window).max() - data.rolling(window).min()) <= self.margin)
            .astype(int)
            .fillna(0)
        )

        # apply backfill if needed: label all points within flatline window
        # as invalid. This requires a forward looking window
        if self.backfill:
            inv_flatliners = flatliners.iloc[::-1]
            inv_flatliners = inv_flatliners.rolling(window, min_periods=1).max()
            flatliners = inv_flatliners.iloc[::-1]

        flatliners = flatliners.astype(bool)

        return flatliners

    def validate(self, X: pd.DataFrame) -> pd.DataFrame:
        """
        Validates the dataframe against the fitted dataframe. Returns a boolean
        dataframe where True indicates an invalid value.

        Parameters
        ----------
        X: pd.DataFrame
            Input dataframe to validate
        """
        invalid_data = pd.DataFrame(
            data=np.zeros_like(X.values).astype(bool),
            index=X.index,
            columns=X.columns,
        )

        for col in self.cols:
            window = self.window_dict[col]
            invalid_data[col] = self._validate_column(X[col], window)

            logger.info(
                f"detected {np.sum(invalid_data[col])} "
                f"flatline samples in {col} "
                f"with window of {window} "
            )

        return invalid_data


class RemoveFlatlines(FlatlineValidator):
    @add_future_warning("RemoveFlatlines is deprecated, use FlatlineValidator instead")
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)



import logging
from dataclasses import dataclass
from typing import List, Optional, Sequence, Tuple, Union, cast

import numpy as np
import pandas as pd
import pytz
from sam.logging_functions import log_dataframe_characteristics, log_new_columns

logger = logging.getLogger(__name__)


@dataclass
class CyclicalMaxes:
    """Class for keeping track of maximum integer values of specific cyclicals"""

    day: int = 31
    dayofweek: int = 7
    weekday: int = 7
    dayofyear: int = 366
    hour: int = 24
    microsecond: int = 1000000
    minute: int = 60
    month: int = 12
    quarter: int = 4
    second: int = 60
    week: int = 53
    secondofday: int = 86400

    @classmethod
    def get_maxes_from_strings(cls, cyclicals: Sequence[str]) -> List[int]:
        """
        This method retrieves cyclical_maxes for pandas datetime features
        The CyclicalMaxes class contains maxes for those features that are actually cyclical.
        For example, 'year' is not cyclical so is not included here.
        Note that the maxes are chosen such that these values are equivalent to 0.
        e.g.: a minute of 60 is equivalent to a minute of 0
        For month, dayofyear and week, these are approximations, but close enough.

        Parameters
        ----------
        cyclicals : list
            List of cyclical strings that match attributes in self class

        Returns
        -------
        list
            List of integer representations of the cyclicals
        """
        for c in cyclicals:
            if c not in cls.__annotations__:
                raise ValueError(
                    str(c) + " is not a known cyclical, please " "provide cyclical_maxes yourself."
                )
        return [getattr(cls, c.lower()) for c in cyclicals]


def decompose_datetime(
    df: pd.DataFrame,
    column: Optional[str] = "TIME",
    components: Optional[Sequence[str]] = None,
    cyclicals: Optional[Sequence[str]] = None,
    onehots: Optional[Sequence[str]] = None,
    remove_categorical: bool = True,
    keep_original: bool = True,
    cyclical_maxes: Optional[Sequence[int]] = None,
    cyclical_mins: Optional[Union[Sequence[int], int]] = (0,),
    timezone: Optional[str] = None,
) -> pd.DataFrame:
    """
    Decomposes a time column to one or more components suitable as features.

    The input is a dataframe with a pandas timestamp column. New columns will be added to this
    dataframe. For example, if column is 'TIME', and components is ['hour', 'minute'], two
    columns: 'TIME_hour' and 'TIME_minute' will be added.

    Optionally, cyclical features can be added instead. For example, if cyclicals is ['hour'],
    then the 'TIME_hour' column will not be added, but two columns 'TIME_hour_sin' and
    'TIME_hour_cos' will be added instead. If you want both the categorical and cyclical features,
    set 'remove_categorical' to False.

    Parameters
    ----------
    df: dataframe
        The dataframe with source column
    column: str (default='TIME')
        Name of the source column to extract components from. Note: time column should have a
        datetime format. if None, it is assumed that the TIME column will be the index.
    components: list
        List of components to extract from datatime column. All default pandas dt components are
        supported, and some custom functions: `['secondofday', 'week']`.
        Note: `week` was added here since it is deprecated in pandas in favor of
        `isocalendar().week`
    cyclicals: list
        List of strings of newly created .dt time variables (like hour, month) you want to convert
        to cyclicals using sine and cosine transformations. Cyclicals are variables that do not
        increase linearly, but wrap around, such as days of the week and hours of the day.
        Format is identical to `components` input.
    onehots: list
        List of strings of newly created .dt time variables (like hour, month) you want to convert
        to one-hot-encoded variables. This is suitable when you think that variables do not
        vary smoothly with time (e.g. Sunday and Monday are quite different).
        This list must be mutually exclusive from cyclicals, i.e. non-overlapping.
    remove_categorical: bool, optional (default=True)
        whether to keep the original cyclical features (i.e. day)
        after conversion (i.e. day_sin, day_cos)
    keep_original: bool, optional (default=True)
        whether to keep the original columns from the dataframe. If this is False, then the
        returned dataframe will only contain newly generated columns, and none of the original ones
    cyclical_maxes: sequence, optional (default=None)
        Passed through to recode_cyclical_features. See :ref:`recode_cyclical_features` for more
        information.
    cyclical_mins: sequence or int, optional (default=0)
        Passed through to recode_cyclical_features. See :ref:`recode_cyclical_features` for more
        information.
    timezone: str, optional (default=None)
        if tz is not None, convert the time to the specified timezone, before creating features.
        timezone can be any string that is recognized by pytz, for example `Europe/Amsterdam`.
        We assume that the TIME column is always in UTC,
        even if the datetime object has no tz info.
    Returns
    -------
    dataframe
        The original dataframe with extra columns containing time components

    Examples
    --------
    >>> from sam.feature_engineering import decompose_datetime
    >>> import pandas as pd
    >>> df = pd.DataFrame({'TIME': pd.date_range("2018-12-27", periods = 4),
    ...                    'OTHER_VALUE': [1, 2, 3, 2]})
    >>> decompose_datetime(df, components= ["year", "dayofweek"])
            TIME  OTHER_VALUE  TIME_year  TIME_dayofweek
    0 2018-12-27            1       2018               3
    1 2018-12-28            2       2018               4
    2 2018-12-29            3       2018               5
    3 2018-12-30            2       2018               6
    """
    components = [] if components is None else components
    cyclicals = [] if cyclicals is None else cyclicals
    onehots = [] if onehots is None else onehots
    cyclical_maxes = [] if cyclical_maxes is None else cyclical_maxes

    if np.any([c in cyclicals for c in onehots]):
        raise ValueError("cyclicals and onehots are not mutually exclusive")

    if keep_original:
        result = df.copy()
    else:
        result = pd.DataFrame(index=df.index)

    if column is None:
        timecol = df.index.to_series().copy()
        column = "" if timecol.name is None else timecol.name
    else:
        timecol = df[column].copy()

    logging.debug(
        f"Decomposing datetime, number of dates: {len(timecol)}. " f"Components: {components}"
    )

    # Fix timezone
    if timezone is not None:
        if timecol.dt.tz is not None:
            if timecol.dt.tz != pytz.utc:
                raise ValueError(
                    "Data should either be in UTC timezone or it should have no"
                    " timezone information (assumed to be in UTC)"
                )
        else:
            timecol = timecol.dt.tz_localize("UTC")
        timecol = timecol.dt.tz_convert(timezone)

    result = _create_time_cols(result, components, timecol, column)

    # do this before converting to cyclicals, as this has its own logging:
    log_new_columns(result, df)
    log_dataframe_characteristics(result, logging.DEBUG)

    # convert cyclicals
    if not isinstance(cyclicals, Sequence):
        raise TypeError("cyclicals must be a sequence type")
    if cyclicals:
        result = recode_cyclical_features(
            result,
            cyclicals,
            prefix=column,
            remove_categorical=remove_categorical,
            cyclical_maxes=cyclical_maxes,
            cyclical_mins=cyclical_mins,
            keep_original=True,
        )
    if onehots:
        result = recode_onehot_features(
            result,
            onehots,
            prefix=column,
            remove_categorical=remove_categorical,
            onehot_maxes=cyclical_maxes,
            onehot_mins=cyclical_mins,
            keep_original=True,
        )

    return result


def recode_cyclical_features(
    df: pd.DataFrame,
    cols: Sequence[str],
    prefix: str = "",
    remove_categorical: bool = True,
    keep_original: bool = True,
    cyclical_maxes: Optional[Sequence[int]] = None,
    cyclical_mins: Optional[Union[Sequence[int], int]] = (0,),
) -> pd.DataFrame:
    """
    Convert cyclical features (like day of week, hour of day) to continuous variables, so that
    Sunday and Monday are close together numerically.

    IMPORTANT NOTE: This function requires a global maximum and minimum for the data. For example,
    for minutes, the global maximum and minimum are 0 and 60 respectively, even if your data never
    reaches these global minimums/maximums explicitly. This function assumes that the minimum and
    maximum should be encoded as the same value: minute 0 and minute 60 mean the same thing.

    If you only use cyclical pandas timefeatures, nothing needs to be done. For these features,
    the minimum/maximum will be chosen automatically. These are: ['day', 'dayofweek', 'weekday',
    'dayofyear', 'hour', 'microsecond', 'minute', 'month', 'quarter', 'second', 'week']

    For any other scenario, global minimums/maximums will need to be passed in the parameters
    `cyclical_maxes` and `cyclical_mins`. Minimums are set to 0 by default, meaning that
    only the maxes need to be chosen as the value that is `equivalent` to 0.

    Parameters
    ----------
    df: pandas dataframe
        Dataframe in which the columns to convert should be present.
    cols: list of strings
        The suffixes column names to convert to continuous numerical values.
        These suffixes will be added to the `column` argument to get the actual column names, with
        a '_' in between.
    column: string, optional (default='')
        name of original time column in df, e.g. TIME.
        By default, assume the columns in cols literally refer to column names in the data
    remove_categorical: bool, optional (default=True)
        whether to keep the original cyclical features (i.e. day)
        after conversion (i.e. day_sin, day_cos)
    keep_original: bool, optional (default=True)
        whether to keep the original columns from the dataframe. If this is False, then the
        returned dataframe will only contain newly generated columns, and none of the original
        ones. If `remove_categorical` is False, the categoricals will be kept, regardless of
        this argument.
    cyclical_maxes: array-like, optional (default=None)
        The maximums that your data can reach. Keep in mind that the maximum value and the
        minimum value will be encoded as the same value. By default, None means that only
        standard pandas timefeatures will be encoded.
    cyclical_mins: array-like or scalar, optional (default=[0])
        The minimums that your data can reach. Keep in mind that the maximum value and the
        minimum value will be encoded as the same value. By default, 0 is used, which is
        applicable for all pandas timefeatures.

    Returns
    -------
    dataframe
        The input dataframe with cols removed, and replaced by the converted features (two for
        each feature).
    """

    new_df, prefix, cyclical_maxes, cyclical_mins = _validate_and_prepare_components(
        df=df,
        cols=cols,
        column=prefix,
        remove_categorical=remove_categorical,
        keep_original=keep_original,
        component_maxes=cyclical_maxes,
        component_mins=cyclical_mins,
    )

    logging.debug(f"Sine/cosine converting cyclicals columns: {cols}")

    for cyclical_min, cyclical_max, col in zip(cyclical_mins, cyclical_maxes, cols):
        if cyclical_min >= cyclical_max:
            raise ValueError(
                "Cyclical min {} is higher than cyclical max {} for column {}".format(
                    cyclical_min, cyclical_max, col
                )
            )

        # prepend column name (like TIME) to match df column names
        col = prefix + col

        if col not in df.columns:
            raise ValueError(f"{col} is not in input dataframe")

        # rescale feature so it runs from 0 to 2*pi:
        # Features that exceed the maximum are rolled over by the sine/cosine:
        # e.g. if min=0 and max=7, 9 will be treated the same as 2
        norm_feature: pd.Series = (df[col] - cyclical_min) / (cyclical_max - cyclical_min)
        norm_feature = 2 * np.pi * norm_feature
        # convert cyclical to 2 variables that are offset:
        new_df[col + "_sin"] = np.sin(norm_feature)
        new_df[col + "_cos"] = np.cos(norm_feature)

        # drop the original. if keep_original is False, this is unneeded: it was already removed
        if remove_categorical and keep_original:
            new_df = new_df.drop(col, axis=1)

    # log changes
    log_new_columns(new_df, df)
    log_dataframe_characteristics(new_df, logging.DEBUG)

    return new_df


def recode_onehot_features(
    df: pd.DataFrame,
    cols: Sequence[str],
    prefix: str = "",
    remove_categorical: bool = True,
    keep_original: bool = True,
    onehot_maxes: Optional[Sequence[int]] = None,
    onehot_mins: Optional[Union[Sequence[int], int]] = (0,),
) -> pd.DataFrame:
    """
    Convert time features (like day of week, hour of day) to onehot variables (1 or 0 for each
    unique value).

    IMPORTANT NOTE: This function requires a global maximum and minimum for the data. For example,
    for minutes, the global maximum and minimum are 0 and 60 respectively, even if your data never
    reaches these global minimums/maximums explicitly. Make sure these variables will all be added
    in your onehot columns, otherwise your columns in train and test set could be unmatching.

    If you only use cyclical pandas timefeatures, nothing needs to be done. For these features,
    the minimum/maximum will be chosen automatically. These are: ['day', 'dayofweek', 'weekday',
    'dayofyear', 'hour', 'microsecond', 'minute', 'month', 'quarter', 'second', 'week']

    For any other scenario, global minimums/maximums will need to be passed in the parameters
    `cyclical_maxes` and `cyclical_mins`. Minimums are set to 0 by default, meaning that
    only the maxes need to be chosen as the value that is `equivalent` to 0.

    Parameters
    ----------
    df: pandas dataframe
        Dataframe in which the columns to convert should be present.
    cols: list of strings
        The suffixes column names to convert to onehot variables.
        These suffixes will be added to the `column` argument to get the actual column names, with
        a '_' in between.
    prefix: string, optional (default='')
        name of original time column in df, e.g. 'TIME'.
        By default, assume the columns in cols literally refer to column names in the data
    remove_categorical: bool, optional (default=True)
        whether to keep the original time features (i.e. day)
    keep_original: bool, optional (default=True)
        whether to keep the original columns from the dataframe. If this is False, then the
        returned dataframe will only contain newly generated columns, and none of the original
        ones. If `remove_categorical` is False, the categoricals will be kept, regardles of
        this argument.
    onehot_maxes: array-like, optional (default=None)
        The maximums that your data can reach. By default, None means that only
        standard pandas timefeatures will be encoded.
    onehot_mins: array-like or scalar, optional (default=[0])
        The minimums that your data can reach. By default, 0 is used, which is
        applicable for all pandas timefeatures.

    Returns
    -------
    pandas dataframe
        The input dataframe with cols removed, and replaced by the converted features.
    """

    new_df, prefix, onehot_maxes, onehot_mins = _validate_and_prepare_components(
        df=df,
        cols=cols,
        column=prefix,
        remove_categorical=remove_categorical,
        keep_original=keep_original,
        component_maxes=onehot_maxes,
        component_mins=onehot_mins,
    )

    logging.debug(f"onehot converting time columns: {cols}")

    for onehot_min, onehot_max, col in zip(onehot_mins, onehot_maxes, cols):
        col = prefix + col

        if col not in df.columns:
            raise ValueError(f"{col} is not in input dataframe")

        # get the onehot encoded dummies
        dummies: pd.DataFrame = pd.get_dummies(df[col], prefix=col).astype(int)

        # fill in the weekdays not in the dataset
        for i in range(onehot_min, onehot_max):
            if not "%s_%d" % (col, i) in dummies.columns:
                dummies["%s_%d" % (col, i)] = 0
        dummies_sorted = dummies[np.sort(dummies.columns)]
        new_df = new_df.join(dummies_sorted)

        # drop the original. if keep_original is False, this is unneeded: it was already removed
        if remove_categorical and keep_original:
            new_df = new_df.drop(col, axis=1)

    log_new_columns(new_df, df)
    log_dataframe_characteristics(new_df, logging.DEBUG)

    return new_df


def _create_time_cols(
    df: pd.DataFrame, components: Sequence[str], timecol: pd.Series, prefix: str = ""
) -> pd.DataFrame:
    """Helper function to create all the neccessary time columns

    Parameters
    ----------
    df : pd.DataFrame
        The input dataframe in SAM format
    components : Sequence[str]
        Time components that need to be added to the data
    timecol: Series
        A pandas series containing the datetimes, used for making the time columns
    prefix : str
        Prefix of the newly created columns, usually the same as the original time column

    Returns
    -------
    pd.DataFrame
        The dataframe, which includes the time components

    Raises
    ------
    NotImplementedError
        In case the time components are not recognized by pandas or by SAM
    """

    pandas_functions = [f for f in dir(timecol.dt) if not f.startswith("_")]

    custom_functions = ["secondofday", "week"]
    for component in components:
        if component in custom_functions:
            if component == "week":
                df[prefix + "_" + component] = timecol.dt.isocalendar().week
            elif component == "secondofday":
                sec_in_min = 60
                sec_in_hour: int = sec_in_min * 60
                df[prefix + "_" + component] = (
                    timecol.dt.hour * sec_in_hour
                    + timecol.dt.minute * sec_in_min
                    + timecol.dt.second
                )
        elif component in pandas_functions:
            df[prefix + "_" + component] = getattr(timecol.dt, component)
        else:
            raise NotImplementedError(f"Component {component} not implemented")
    return df


def _validate_and_prepare_components(
    df: pd.DataFrame,
    cols: Sequence[str],
    column: str,
    remove_categorical: bool,
    keep_original: bool,
    component_maxes: Optional[Sequence[int]],
    component_mins: Optional[Union[Sequence[int], int]],
) -> Tuple[pd.DataFrame, str, Sequence[int], Sequence[int]]:
    """
    Validates and prepares the dataframe, component (onehot or cyclical) parameters and min/max
    component bounds.

    Parameters
    ----------
    df: pandas dataframe
        Dataframe in which the columns to convert should be present.
    cols: list of strings
        The suffixes column names to convert to onehot variables.
        These suffixes will be added to the `column` argument to get the actual column names, with
        a '_' in between.
    column: string
        name of original time column in df (e.g. TIME)
        By default, assume the columns in cols literally refer to column names in the data
    remove_categorical: bool
        whether to keep the original time features (i.e. day)
    keep_original: bool
        whether to keep the original columns from the dataframe. If this is False, then the
        returned dataframe will only contain newly generated columns, and none of the original
        ones. If `remove_categorical` is False, the categoricals will be kept, regardles of
        this argument.
    component_maxes: array-like
        The maximums that your data can reach. By default, None means that only
        standard pandas timefeatures will be encoded.
    component_mins: array-like or scalar
        The minimums that your data can reach. By default, 0 is used, which is
        applicable for all pandas timefeatures.

    Returns
    -------
    new_df: pandas dataframe
        Dataframe with/without the categoricals and the original columns depending on setttings.
    column: string
        updated name of original time column in df, e.g. 'TIME_'.
    component_maxes: array-like
        The maximums that your data can reach.
    component_mins: array-like
        The minimums that your data can reach.
    """

    column = column + "_" if column != "" else column
    if not isinstance(df, pd.DataFrame):
        raise TypeError("df should be pandas dataframe")
    if not isinstance(cols, Sequence):
        raise TypeError("cols should be a sequence of columns to convert")
    if not isinstance(remove_categorical, bool):
        raise TypeError("remove_categorical should be a boolean")

    if component_maxes is None or not component_maxes:
        component_maxes = CyclicalMaxes.get_maxes_from_strings(cols)

    if np.isscalar(component_mins):
        component_mins = [cast(int, component_mins)] * len(component_maxes)
    elif isinstance(component_mins, (Sequence, np.ndarray)):
        if len(component_mins) == 1:
            component_mins = list(component_mins)
            component_mins *= len(component_maxes)
    else:
        raise TypeError("`component_maxes` needs to be a scalar or array-like")

    if keep_original:
        new_df = df.copy()
    elif not remove_categorical:
        # We don't want to keep the original columns, but we want to keep the categoricals
        new_df = df.copy()[[column + col for col in cols]]
    else:
        new_df = pd.DataFrame(index=df.index)

    return new_df, column, component_maxes, component_mins



from .diagnostic_flatline_removal import diagnostic_flatline_removal  # noqa: F401
from .extreme_removal_plot import diagnostic_extreme_removal
from .incident_heatmap import plot_incident_heatmap
from .performance_evaluation_fixed_predict_ahead import (
    performance_evaluation_fixed_predict_ahead,
)
from .plot_feature_importances import plot_feature_importances
from .precision_recall import plot_precision_recall_curve
from .quantile_plot import sam_quantile_plot
from .rolling_correlations import plot_lag_correlation
from .threshold_plot import plot_threshold_curve

__all__ = [
    "plot_precision_recall_curve",
    "plot_threshold_curve",
    "plot_incident_heatmap",
    "plot_lag_correlation",
    "diagnostic_extreme_removal",
    "sam_quantile_plot",
    "performance_evaluation_fixed_predict_ahead",
    "plot_feature_importances",
]



from .decompose_datetime import decompose_datetime, recode_cyclical_features
from .rolling_features import BuildRollingFeatures
from .lag_range import range_lag_column
from .automatic_rolling_engineering import AutomaticRollingEngineering
from .weather_spei import SPEITransformer
from .base_feature_engineering import BaseFeatureEngineer, FeatureEngineer, IdentityFeatureEngineer
from .simple_feature_engineering import SimpleFeatureEngineer


__all__ = [
    "decompose_datetime",
    "recode_cyclical_features",
    "BuildRollingFeatures",
    "range_lag_column",
    "AutomaticRollingEngineering",
    "SPEITransformer",
    "BaseFeatureEngineer",
    "FeatureEngineer",
    "IdentityFeatureEngineer",
    "SimpleFeatureEngineer",
]



import warnings
from typing import List

try:
    # Tensorflow often raises warnings when importing.
    # When importing sam, it is not necessary to show these warnings since they
    # are not relevant to sam
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", category=DeprecationWarning)
        warnings.simplefilter("ignore", category=FutureWarning)
        import tensorflow as tf
        import tensorflow.keras.backend as K
except ImportError:
    # These are optional dependencies so it's not necessary to crash if they aren't found.
    # However, this will crash once one of the functions below runs.
    pass


def keras_tilted_loss(y_true: tf.Tensor, y_pred: tf.Tensor, quantile: float = 0.5):
    """
    Calculate tilted, or quantile loss in Keras. Given a quantile q, and an error e,
    tilted loss is defined as `(1-q) * |e|` if `e < 0`, and `q * |e|` if `e > 0`.

    This function is the same as the mean absolute error if q=0.5, which approximates the median.
    For a given value of q, the function that minimizes the tilted loss will be the q'th quantile
    function.

    Parameters
    ----------
    y_true: Theano/TensorFlow tensor.
        True labels.
    y_pred: Theano/TensorFlow tensor.
        Predictions. Must be same shape as `y_true`
    quantile: float, optional (default=0.5)
        The quantile to use when computing tilted loss.
        Must be between 0 and 1 for tilted loss to be positive.

    Examples
    --------
    >>> from sam.metrics import keras_tilted_loss
    >>> from sam.models import create_keras_quantile_mlp
    >>> n_input = 1000
    >>> n_neurons = 64
    >>> n_layers = 3
    >>> quantiles = [0.1, 0.5, 0.9]
    >>> model = create_keras_quantile_mlp(n_input, n_neurons, n_layers, quantiles=quantiles)
    >>> quantile = 0.5  # Quantile, in this case the median
    >>> model.compile(loss=lambda y,f: keras_tilted_loss(y, f, quantile))
    """
    e = y_true - y_pred
    return K.mean(K.maximum(quantile * e, (quantile - 1) * e), axis=-1)


def keras_joint_mse_tilted_loss(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    quantiles: List[float] = None,
    n_targets: int = 1,
):
    """
    Joint mean and quantile regression loss function using mse.
    Sum of mean squared error and multiple tilted loss functions
    Custom loss function, inspired by https://github.com/fmpr/DeepJMQR
    Only compatible with tensorflow backend.

    This calculates loss for multiple quantiles, and multiple targets.
    The total loss is the sum of the mse of all targets, and the tilted
    loss of all quantiles.

    `y_true` is expected to be a tensor with shape `(None, n_targets)`
    `y_pred` is expected to be a tensor with shape `(None, n_targets * (n_quantiles + 1))`
    For example, if there are 2 outputs and 2 quantiles, the order of y_pred
    should be: `[output_1_quantile_1, output_2_quantile_1,
    output_1_quantile_2, output_2_quantile_2, output_1_mean, output_2_mean]`

    Parameters
    ----------
    y_true: tensorflow tensor
        True values
    y_pred: tensorflow tensor
        Predicted values
    quantiles: list of floats (default=None)
        Quantiles to predict. Values between 0 and 1. By default, no quantile loss is used.
    n_targets: integer, optional (default=1)
        The number of distinct outputs to predict

    Examples
    --------
    >>> from sam.metrics import keras_joint_mse_tilted_loss as mse_tilted
    >>> from sam.models import create_keras_quantile_mlp
    >>> n_input = 1000
    >>> n_neurons = 64
    >>> n_layers = 3
    >>> quantiles = [0.1, 0.5, 0.9]
    >>> model = create_keras_quantile_mlp(n_input, n_neurons, n_layers, quantiles=quantiles)
    >>> qs = [0.1, 0.9]
    >>> model.compile(loss=lambda y,f: mse_tilted(y, f, qs))
    """
    if quantiles is None:
        quantiles = []
    # select the last column (nodes) of the output
    k = len(quantiles)
    mean_pred = tf.slice(y_pred, [0, k * n_targets], [-1, n_targets])
    # The last node will be fit with regular mean squared error
    loss = K.sum(K.mean(K.square(y_true - mean_pred), axis=0), axis=-1)
    # For each quantile fit one node with corresponding tilted loss
    for k in range(len(quantiles)):
        q = quantiles[k]
        # Select the kth node
        q_pred = tf.slice(y_pred, [0, k * n_targets], [-1, n_targets])
        e = y_true - q_pred
        # add tilted loss to total loss
        loss += K.sum(K.mean(K.maximum(q * e, (q - 1) * e), axis=0), axis=-1)
    return loss


def keras_joint_mae_tilted_loss(
    y_true: tf.Tensor,
    y_pred: tf.Tensor,
    quantiles: List[float] = None,
    n_targets: int = 1,
):
    """Joint mean and quantile regression loss function using mae.
    Sum of mean absolute error and multiple tilted loss functions
    Custom loss function, inspired by https://github.com/fmpr/DeepJMQR
    Only compatible with tensorflow backend.

    This calculates loss for multiple quantiles, and multiple targets.
    The total loss is the sum of the mae of all targets, and the tilted
    loss of all quantiles.

    ``y_true`` is expected to be a tensor with shape ``(None, n_targets)``
    ``y_pred`` is expected to be a tensor with shape ``(None, n_targets * (n_quantiles + 1))``
    For example, if there are 2 outputs and 2 quantiles, the order of y_pred
    should be: `[output_1_quantile_1, output_2_quantile_1,
    output_1_quantile_2, output_2_quantile_2, output_1_mean, output_2_mean]`

    Parameters
    ----------
    y_true: tensorflow tensor
        True values
    y_pred: tensorflow tensor
        Predicted values
    quantiles: list of floats (default=None)
        Quantiles to predict. Values between 0 and 1. By default, no quantile loss is used.
    n_targets: integer, optional (default=1)
        The number of distinct outputs to predict

    Examples
    --------
    >>> from sam.metrics import keras_joint_mae_tilted_loss as mae_tilted
    >>> from sam.models import create_keras_quantile_mlp
    >>> n_input = 1000
    >>> n_neurons = 64
    >>> n_layers = 3
    >>> quantiles = [0.1, 0.5, 0.9]
    >>> model = create_keras_quantile_mlp(n_input, n_neurons, n_layers, quantiles=quantiles)
    >>> qs = [0.1, 0.9]
    >>> model.compile(loss=lambda y,f: mae_tilted(y, f, qs))
    """
    if quantiles is None:
        quantiles = []
    # select the last column (nodes) of the output
    k = len(quantiles)
    mean_pred = tf.slice(y_pred, [0, k * n_targets], [-1, n_targets])
    # The last node will be fit with 0.5 quantile
    loss = K.sum(K.mean(K.abs(y_true - mean_pred), axis=0), axis=-1)
    # For each quantile fit one node with corresponding tilted loss
    for k in range(len(quantiles)):
        q = quantiles[k]
        # Select the kth node
        q_pred = tf.slice(y_pred, [0, k * n_targets], [-1, n_targets])
        e = y_true - q_pred
        # add tilted loss to total loss
        loss += K.sum(K.mean(K.maximum(q * e, (q - 1) * e), axis=0), axis=-1)
    return loss


def keras_rmse(y_true: tf.Tensor, y_pred: tf.Tensor):
    """
    Calculate root mean squared error in Keras.

    Parameters
    ----------
    y_true: Theano/TensorFlow tensor.
        True labels.
    y_pred: Theano/TensorFlow tensor.
        Predictions. Must be same shape as `y_true`

    Examples
    --------
    >>> from sam.metrics import keras_rmse
    >>> from sam.models import create_keras_quantile_mlp
    >>> n_input = 1000
    >>> n_neurons = 64
    >>> n_layers = 3
    >>> quantiles = [0.1, 0.5, 0.9]
    >>> model = create_keras_quantile_mlp(n_input, n_neurons, n_layers, quantiles=quantiles)
    >>> model.compile(loss=keras_rmse)
    """
    return K.sqrt(K.mean(K.square(y_pred - y_true), axis=-1))


def get_keras_forecasting_metrics(quantiles: List[float] = None):
    """
    Get list of standard forecasting metrics.
    These metrics could be used specifically for regression problems, such as forecasting.
    The default list is mse, mae, rmse. Quantile loss can also be added, by giving
    a list of quantiles.

    Parameters
    ----------
    quantiles: list of floats (default=None)
        List of quantiles to make losses of. Values between 0 and 1.
        By default, no quantile loss is used. Note that mae is already included and is
        the same as quantile loss with quantile 0.5.

    Examples
    --------
    >>> from sam.metrics import keras_rmse, get_keras_forecasting_metrics
    >>> from sam.models import create_keras_quantile_mlp
    >>> n_input = 1000
    >>> n_neurons = 64
    >>> n_layers = 3
    >>> quantiles = [0.1, 0.5, 0.9]
    >>> model = create_keras_quantile_mlp(n_input, n_neurons, n_layers, quantiles=quantiles)
    >>> model.compile(loss=keras_rmse, metrics=get_keras_forecasting_metrics())
    """
    if quantiles is None:
        quantiles = []
    quantile_metrics = [lambda y, m: keras_tilted_loss(y, m, q) for q in quantiles]
    main_metrics = ["mse", "mae", keras_rmse]
    return main_metrics + quantile_metrics



import warnings
from abc import ABC, abstractmethod
from operator import itemgetter
from typing import Callable, List, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from sam.feature_engineering import BaseFeatureEngineer, IdentityFeatureEngineer
from sam.models.utils import remove_target_nan, remove_until_first_value
from sam.metrics import joint_mae_tilted_loss, joint_mse_tilted_loss
from sam.preprocessing import inverse_differenced_target, make_shifted_target
from sam.utils import assert_contains_nans, make_df_monotonic
from sklearn.base import BaseEstimator, RegressorMixin, TransformerMixin
from sklearn.utils.validation import check_is_fitted
from sklearn.pipeline import Pipeline


class BaseTimeseriesRegressor(BaseEstimator, RegressorMixin, ABC):
    """
    This is an abstract class for all SAM models.
    Every SAM model (including the most used MLPTimeseriesRegressor) needs to inherit this class
    and implement the abstract methods.

    There are some notes:
    - There is no validation yet. Therefore, the input data must already be sorted and monospaced
    - The feature engineering should be provided as a any transformer. If the feature engineering
      is not provided, the identity transformer will be used. Good practice is for example to use
      the SimpleFeatureEngineer
    - The prediction requires y as input. The reason for this is described in the predict function.
      Keep in mind that this is not directly 'cheating', since we are predicting a future value of
      y, and giving the present value of y as input to the predict.
      When predicting the present, this y is not needed and can be None

    Note that the below parameters are just for the abstract base class, and subclasses can have
    different `__init__` parameters.

    Parameters
    ----------
    predict_ahead: tuple of integers, optional (default=(0,))
        how many steps to predict ahead. For example, if (1, 2), the model will predict both 1 and
        2 timesteps into the future. If (0,), predict the present.
    quantiles: tuple of floats, optional (default=())
        The quantiles to predict. Values between 0 and 1. Keep in mind that the mean will be
        predicted regardless of this parameter
    use_diff_of_y: bool, optional (default=False)
        If True differencing is used (the difference between y now and shifted y),
        else differencing is not used (shifted y is used).
    timecol: string, optional (default=None)
        If not None, the column to use for constructing time features. For now,
        creating features from a DateTimeIndex is not supported yet.
    y_scaler: object, optional (default=None)
        Should be an sklearn-type transformer that has a transform and inverse_transform method.
        E.g.: StandardScaler() or PowerTransformer()
    average_type: str (default='mean')
        Determines what to fit as the average: 'mean', or 'median'. The average is the last
        node in the output layer and does not reflect a quantile, but rather estimates the central
        tendency of the data. Setting to 'mean' results in fitting that node with MSE, and
        setting this to 'median' results in fitting that node with MAE (equal to 0.5 quantile).
    feature_engineering: object, optional (default=None)
        Should be an sklearn-type transformer that has a transform method, e.g.
        `sam.feature_engineering.SimpleFeatureEngineer`.
    kwargs: dict, optional
        Not used. Just for compatibility of models that inherit from this class.

    Attributes
    ----------
    feature_engineer_: Sklearn transformer
        The transformer used on the raw data before prediction
    prediction_cols_: array of strings
        The names of the output columns from the model.
    model_: underlying model (not implemented here)
        The underlying model
    """

    def __init__(
        self,
        predict_ahead: Sequence[int] = (0,),
        quantiles: Sequence[float] = (),
        use_diff_of_y: bool = False,
        timecol: str = None,
        y_scaler: TransformerMixin = None,
        average_type: str = "mean",
        feature_engineer: BaseFeatureEngineer = None,
        **kwargs,
    ) -> None:
        self.predict_ahead = predict_ahead
        self.quantiles = quantiles
        self.use_diff_of_y = use_diff_of_y
        self.timecol = timecol
        self.y_scaler = y_scaler
        self.average_type = average_type
        self.feature_engineer_ = (
            feature_engineer if feature_engineer else IdentityFeatureEngineer()
        )

    @abstractmethod
    def get_untrained_model(self) -> Callable:
        """Returns an underlying model that can be trained

        This abstract method needs to be implemented by any class that inherits from
        BaseTimeseriesRegressor. It returns a trainable model like Sklearn, keras or
        anything you can think of

        Returns
        -------
        A trainable model class
        """
        raise NotImplementedError("Abstract method. Needs to be implemented by subclass")

    def validate_predict_ahead(self):
        """
        Perform checks to validate the predict_ahead attribute
        """
        if np.isscalar(self.predict_ahead):
            self.predict_ahead = [self.predict_ahead]

        if not all([p >= 0 for p in self.predict_ahead]):
            raise ValueError("All values of predict_ahead must be 0 or larger!")

        if 0 in self.predict_ahead and self.use_diff_of_y:
            raise ValueError("use_diff_of_y must be false when predict_ahead is 0")

        if len(np.unique(self.predict_ahead)) != len(self.predict_ahead):
            raise ValueError("predict_ahead contains double values")

    def validate_data(self, X: pd.DataFrame) -> None:
        """
        Validates the data and raises an exception if:
        - There is no time columns
        - The data is not monospaced

        Parameters
        ----------
        x: pd.DataFrame
            The dataframe to validate
        """
        if self.timecol is None:
            if isinstance(X.index, pd.DatetimeIndex):
                monospaced = X.index.to_series().diff().dropna().unique().size == 1
            else:
                warnings.warn(
                    (
                        "No timecolumn given. Make sure the data is"
                        "monospaced when given to this model!"
                    ),
                    UserWarning,
                )
                monospaced = True
        else:
            monospaced = X[self.timecol].diff()[1:].unique().size == 1
        if not monospaced:
            raise ValueError(
                "Data is not monospaced, which is required for"
                "this model. fit/predict is not possible"
            )

    @staticmethod
    def verify_same_indexes(X: pd.DataFrame, y: pd.Series, y_can_be_none=True):
        """
        Verify that X and y have the same index
        """
        if not y.index.equals(X.index):
            raise ValueError("For training, X and y must have an identical index")

    def preprocess(self, X: pd.DataFrame, y: pd.DataFrame, train: bool = False):
        """
        Preprocess the data. This is the first step in the pipeline.
        """
        X, y = X.copy(), y.copy()
        y = make_shifted_target(y=y, use_diff_of_y=self.use_diff_of_y, lags=self.predict_ahead)
        if train:
            if self.y_scaler is not None:
                y = pd.DataFrame(
                    self.y_scaler.fit_transform(y),
                    index=X.index,
                    columns=y.columns,
                )
            self._set_input_cols(X)
            X = pd.DataFrame(
                self.feature_engineer_.fit_transform(X),
                index=X.index,
                columns=self.get_feature_names_out(),
            )
            self.n_inputs_ = len(self.get_feature_names_out())
        else:
            if self.y_scaler is not None:
                y = pd.DataFrame(
                    self.y_scaler.transform(y),
                    index=X.index,
                    columns=y.columns,
                )
            X = pd.DataFrame(
                self.feature_engineer_.transform(X),
                index=X.index,
                columns=self.get_feature_names_out(),
            )

        X, y = remove_until_first_value(X, y)
        X, y = remove_target_nan(X, y, use_x=False)

        return X, y

    def preprocess_fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        validation_data: Tuple[pd.DataFrame, pd.Series] = None,
    ) -> Tuple[pd.DataFrame, pd.Series, pd.DataFrame, pd.Series]:
        """
        This function does the following:
        - Validate that the input is monospaced and has enough rows
        - Perform differencing on the target
        - Fitting/applying the feature engineer
        - Bookkeeping to create the output columns
        - Remove rows with nan that can't be used for fitting
        - Optionally, preprocess validation data to give to the fit

        Parameters
        ----------
        X: pd.DataFrame
            The independent variables used to 'train' the model
        y: pd.Series
            Target data (dependent variable) used to 'train' the model.
        validation_data: tuple(pd.DataFrame, pd.Series) (X_val, y_val respectively)
            Data used for validation step

        Returns
        -------
        X_transformed: pd.DataFrame
            The transformed feature table, ready to be used in fitting the model
        y_transformed: pd.Series
            The transformed target, ready to be used in fitting the model
        X_val_transformed: pd.DataFrame
            The transformed feature table, ready to be used for validating the model
            If no validation data is provided, this returns None
        y_val_transformed: pd.Series
            The transformed target, ready to be used for validating the model
            If no validation data is provided, this returns None
        """
        BaseTimeseriesRegressor.verify_same_indexes(X, y)
        self.validate_predict_ahead()
        self.validate_data(X)

        # Create output column names. In this model, our outputs are assumed to have the
        # form: [quantile_1_output_1, quantile_1_output_2, ... ,
        # quantile_n_output_1, quantile_n_output_2, ..., mean_output_1, mean_output_2]
        # Where n_output (1 or 2 in this example) is decided by self.predict_ahead
        self.prediction_cols_ = [
            "predict_lead_{}_q_{}".format(p, q) for q in self.quantiles for p in self.predict_ahead
        ]
        self.prediction_cols_ += ["predict_lead_{}_mean".format(p) for p in self.predict_ahead]
        self.n_outputs_ = len(self.prediction_cols_)

        X_transformed, y_transformed = self.preprocess(X, y, train=True)

        assert_contains_nans(
            X_transformed, "Data cannot contain nans. Imputation not supported for now"
        )

        # Apply transformations to validation data if provided:
        if validation_data is not None:
            X_val, y_val = validation_data
            self.validate_data(X_val)
            X_val_transformed, y_val_transformed = self.preprocess(X_val, y_val, train=False)
        else:
            X_val_transformed, y_val_transformed = None, None

        return X_transformed, y_transformed, X_val_transformed, y_val_transformed

    @abstractmethod
    def fit(
        self,
        X: pd.DataFrame,
        y: pd.Series,
        validation_data: Tuple[pd.DataFrame, pd.Series] = None,
        **fit_kwargs,
    ) -> Callable:
        """Fit the underlying model

        This abstract method needs to be implemented by any class inheriting from
        BaseTimeseriesRegressor. This function receives preprocessed input data and
        applies feature building and trains the underlying model.

        It is advised to use the `preprocess_fit()` function in any implementation.

        Parameters
        ----------
        X : pd.DataFrame
            The preprocessed input data for the model
        y : pd.Series
            The preprocessed target for the model
        validation_data : Tuple[pd.DataFrame, pd.Series], optional
            Validation data to calculate metric scores during training, by default None

        Returns
        -------
        Callable
            Usually this returns the history of the fit (when using keras)
        """
        raise NotImplementedError("Abstract method. Needs to be implemented by subclass")

    @abstractmethod
    def predict(
        self, X: pd.DataFrame, y: pd.Series = None, return_data: bool = False
    ) -> Union[pd.DataFrame, Tuple[pd.DataFrame, pd.DataFrame]]:
        """Predict on new data using a trained model

        This abstract method needs to be implemented by any class inheriting from
        BaseTimeseriesRegressor. This function receives preprocessed input data and
        applies feature building and predicts the target using a trained model.

        Important! This is different from sklearn/tensorflow API...
        We need y during prediction for two reasons:
        1) a lagged version is used for feature engineering
        2) The underlying model can predict a differenced number, and then we want to output the
           'real' prediction, so we need y to undo the differencing
        Keep in mind that prediction will work if you are predicting the future. e.g. you have
        data from 00:00-12:00, and are predicting 4 hours into the future, it will predict what
        the value will be at 4:00-16:00

        It is advised to use the `preprocess_predict()` and
        `postprocess_predict()` functions in any implementation.

        Parameters
        ----------
        X : pd.DataFrame
            The independent variables used to predict.
        y : pd.Series, optional
            The target values, by default None
        return_data : bool, optional
            whether to return only the prediction, or to return both the prediction and the
            transformed input (X) dataframe., by default False

        Returns
        -------
        prediction: pd.DataFrame
            The predictions coming from the model
        X_transformed: pd.DataFrame, optional
            The transformed input data, when return_data is True, otherwise None
        """
        raise NotImplementedError("Abstract method. Needs to be implemented by subclass")

    def preprocess_predict(
        self, X: pd.DataFrame, y: pd.Series, dropna: bool = False
    ) -> pd.DataFrame:
        """
        Transform a DataFrame X so it can be fed to self.model_.
        This is useful for several usecases, where you want to use the underlying
        keras model as opposed to the wrapper. For example shap, eli5, and even just
        implementing the `predict` function.

        Parameters
        ----------
        X: pd.DataFrame
            The training input samples.
        y: pd.Series
            The target values.
        dropna: bool, optional (default=False)
            If True, delete the rows that contain NaN values.
        """
        # This function only works if the estimator is fitted
        check_is_fitted(self, "feature_engineer_")

        if y is not None:
            BaseTimeseriesRegressor.verify_same_indexes(X, y)

        X_transformed = pd.DataFrame(
            self.feature_engineer_.transform(X),
            index=X.index,
            columns=self.get_feature_names_out(),
        )

        if dropna:
            X_transformed = X_transformed[~np.isnan(X_transformed).any(axis=1)]
        return X_transformed

    def postprocess_predict(
        self,
        prediction: pd.DataFrame,
        X: pd.DataFrame,
        y: pd.Series,
        force_monotonic_quantiles: bool = False,
    ) -> pd.DataFrame:
        """
        Postprocessing function for the prediction result.

        Parameters
        ----------
        prediction : pd.DataFrame
            Dataframe containing the prediction.
        X : pd.DataFrame
            The training input samples.
        y : pd.Series
            The target values.
        force_monotonic_quantiles : bool, optional (default=False)
            whether to force quantiles to not overlap. When fitting multiple quantile regressions
            it is possible that individual quantile regression lines over-lap, or in other words,
            a quantile regression line fitted to a lower quantile predicts higher that a line
            fitted to a higher quantile. If this occurs for a certain prediction, the output
            distribution is invalid. In this function we force monotonicity by making the outer
            quantiles at least as high as the inner quantiles.

        Returns
        -------
        pd.DataFrame
            Postprocessed predictions.
        """
        # Put the predictions in a dataframe so we can undo the differencing
        prediction = pd.DataFrame(prediction, columns=self.prediction_cols_, index=X.index)

        # Undo the scaling per quantile
        if self.y_scaler is not None:
            for output_col in [f"q_{q}" for q in self.quantiles] + ["mean"]:
                mask = prediction.columns.str.contains(output_col)
                sub_prediction = prediction.loc[:, mask]
                sub_prediction = pd.DataFrame(
                    self.y_scaler.inverse_transform(sub_prediction.values),
                    index=sub_prediction.index,
                    columns=sub_prediction.columns,
                )
                prediction.loc[:, mask] = sub_prediction

        # Undo the differencing
        if self.use_diff_of_y:
            prediction = inverse_differenced_target(prediction, y)

        if force_monotonic_quantiles:
            prediction = self.make_prediction_monotonic(prediction)

        if prediction.shape[1] == 1:
            prediction = prediction.iloc[:, 0]

        return prediction

    def make_prediction_monotonic(self, prediction: pd.DataFrame) -> pd.DataFrame:
        """
        When fitting multiple quantile regressions it is possible that individual quantile
        regression lines over-lap, or in other words, a quantile regression line fitted to a lower
        quantile predicts higher that a line fitted to a higher quantile. If this occurs for a
        certain prediction, the output distribution is invalid. In this function we force
        monotonicity by making the outer quantiles at least as high as the inner quantiles.

        Parameters
        ----------
        prediction : pd.DataFrame
            Dataframe containing a prediction containing several quantiles

        Returns
        -------
        pd.DataFrame
            Prediction for which the quantiles are (now) monotonic
        """
        # Retrieve prediction columns and split them by group (predict ahead)
        grouped_cols = {}
        for p in self.predict_ahead:
            grouped_cols[p] = [
                [q, "predict_lead_{}_q_{}".format(p, q)]
                for q in self.quantiles
                if "predict_lead_{}_q_{}".format(p, q) in prediction.columns
            ]

        # Divide and order ascending
        lower_band_groups = []
        upper_band_groups = []
        for predict_ahead in grouped_cols:
            sorted_cols = grouped_cols[predict_ahead]
            sorted_cols.sort(key=itemgetter(0))
            upper_band = [x[1] for x in sorted_cols if x[0] > 0.5]
            lower_band = [x[1] for x in sorted_cols if x[0] < 0.5][::-1]
            if upper_band:
                upper_band_groups.append(upper_band)
            if lower_band:
                lower_band_groups.append(lower_band)

        # Upper band quantiles should monotonic increase, and lower band quantiles should
        # monotonic decrease
        for g in upper_band_groups:
            prediction[g] = make_df_monotonic(prediction[g], aggregate_func="max")
        for g in lower_band_groups:
            prediction[g] = make_df_monotonic(prediction[g], aggregate_func="min")

        return prediction

    def get_feature_names_out(self, input_features=None) -> List[str]:
        """
        Function for obtaining feature names. Generally used instead of the attribute, and more
        compatible with the sklearn API.

        Returns
        -------
        list:
            list of feature names
        """

        if hasattr(self.feature_engineer_, "feature_engineer_"):
            return self.feature_engineer_.get_feature_names_out()
        elif isinstance(self.feature_engineer_, Pipeline):
            # select last step that has get_feature_names method
            feature_steps = [
                step[-1]
                for step in self.feature_engineer_.steps
                if hasattr(step[-1], "get_feature_names_out")
            ]
            if len(feature_steps) > 0:
                return feature_steps[-1].get_feature_names_out()
            else:
                raise ValueError(
                    "Feature engineering pipelines require at least one step "
                    "with get_feature_names method"
                )
        else:
            raise ValueError("Feature engineering must be a Pipeline or a BaseFeatureEngineer")

    def _set_input_cols(self, X: pd.DataFrame) -> None:
        """
        Function to set the attribute self._input_cols (input column names).
        Only used internally right before the feature building.
        Time column is not included, since time is always a dependency
        This can be used to determine model dependencies

        Parameters
        ----------
        X: pd.DataFrame
            The DataFrame that contains the input columns
        """
        col_names = X.columns.values
        col_names = col_names[col_names != self.timecol]
        self._input_cols = col_names

    def get_input_cols(self) -> np.ndarray:
        """
        Function to obtain the input column names.
        This can be used to determine model dependencies
        Time column is not included, since time is always a dependency

        Returns
        -------
        list:
            The input column names
        """
        check_is_fitted(self, "_input_cols")
        return self._input_cols

    def get_actual(self, y: pd.Series) -> Union[pd.Series, pd.DataFrame]:
        """
        Convenience function for getting the actual values (perfect prediction).
        Mainly useful for scoring the model. This essentially does and undoes differencing
        on y, meaning this function will output what a perfect model would have outputted.
        If predict_ahead is 0, no differencing is done anyway, so y is just returned unchanged.

        Returns a Series of a Dataframe depending on the number of values in self.predict_ahead

        Parameters
        -------
        y:
            The target values

        Returns
        -------
        pd.Series or pd.DataFrame:
            y after applying differencing and undoing the process (if self.use_diff_of_y)
            If self.predict_ahead is a single value, this function will return a series.
            If self.predict_ahead has multiple values (list), this function will return a
            dataframe.
        """
        # This function only works if the estimator is fitted
        check_is_fitted(self, "model_")

        # No scaling or differencing applied, because the predict() method already reversed this
        actual = make_shifted_target(y=y, use_diff_of_y=False, lags=self.predict_ahead)

        if actual.shape[1] == 1:
            actual = actual.iloc[:, 0]

        return actual

    def score(self, X: pd.DataFrame, y: pd.Series) -> float:
        """
        Default score function. Uses sum of tilted loss of quantile predictions plus the mse
        of the mean predictions or mae of the median predictions.

        Parameters
        ----------
        X: pd.DataFrame
            The independent variables used to predict.
        y: pd.Series
            The target values

        Returns
        -------
        float:
            The score
        """
        # This function only works if the estimator is fitted
        check_is_fitted(self, "model_")
        # We need a dataframe, regardless of if these functions outputs a series or dataframe
        prediction = pd.DataFrame(self.predict(X, y), columns=self.prediction_cols_)
        actual = pd.DataFrame(self.get_actual(y))

        # scale these predictions back to get a score that is in same units as keras loss
        if self.y_scaler is not None:
            pred_scaled = np.zeros_like(prediction)
            for i in range(prediction.shape[1]):
                pred_scaled[:, i] = self.y_scaler.transform(
                    prediction.iloc[:, i].values.reshape(-1, 1)
                ).ravel()
            prediction = pd.DataFrame(
                pred_scaled, columns=prediction.columns, index=prediction.index
            )

            actual = pd.DataFrame(
                self.y_scaler.transform(actual.values).ravel(),
                index=actual.index,
                columns=actual.columns,
            )

        # actual usually has some missings at the end
        # prediction usually has some missings at the beginning
        # We ignore the rows with missings
        prediction, actual = remove_target_nan(prediction, actual, use_x=True)

        # Calculate the joint tilted loss of all the average and quantile predictions
        if self.average_type == "median":
            loss = joint_mae_tilted_loss(
                actual, prediction, quantiles=self.quantiles, n_targets=len(self.predict_ahead)
            )
        elif self.average_type == "mean":
            loss = joint_mse_tilted_loss(
                actual, prediction, quantiles=self.quantiles, n_targets=len(self.predict_ahead)
            )

        score = -1 * loss
        return score

    @abstractmethod
    def dump(self, foldername: str, prefix: str = "model") -> None:
        """Save a model to disk

        This abstract method needs to be implemented by any class inheriting from
        BaseTimeseriesRegressor. This function dumps the SAM model to disk.

        Parameters
        ----------
        foldername : str
            The folder location where to save the model
        prefix : str, optional
           The prefix used in the filename, by default "model"
        """
        return None

    @classmethod
    @abstractmethod
    def load(cls, foldername, prefix="model"):
        """Load a model from disk

        This abstract method needs to be implemented by any class inheriting from
        BaseTimeseriesRegressor. This function loads a SAM model from disk.

        Parameters
        ----------
        foldername : str
            The folder location where the model is stored
        prefix : str, optional
           The prefix used in the filename, by default "model"

        Returns
        -------
        The SAM model that has been loaded from disk
        """
        return None
